from __future__ import annotations

from collections.abc import Iterable
import datetime
import enum
import logging
import os
import pathlib

# from typing import TypedDict
import fsspec

from prettyqt import constants, core, gui, itemmodels, widgets
from prettyqt.utils import datatypes


logger = logging.getLogger(__name__)


class FsSpecCompleter(widgets.Completer):
    def pathFromIndex(self, index):
        return index.data(self.model().Roles.FilePathRole)

    def splitPath(self, path: str):
        return path.split(self.model().fs.sep)


# class FolderInfo(TypedDict):
#     name: str
#     size: int
#     type: str
#     created: float
#     islink: bool  # symbolic link
#     mode: int  # Inode protection mode.
#     uid: int  # user id of owner
#     gid: int  # group id of owner
#     mtime: float
#     ino: int  # Inode number.
#     nlink: int  # Number of links to the inode.


_icon_provider = widgets.FileIconProvider()


VALUE_LETTERS = [(4, "r"), (2, "w"), (1, "x")]


class FsSpecColumnItem(itemmodels.ColumnItem):
    identifier: str = ""

    def get_data(self, item: FSSpecTreeModel.TreeItem, role: constants.ItemDataRole):
        name = item.obj["name"]
        match role:
            case FSSpecTreeModel.Roles.FilePathRole:
                return name
            case FSSpecTreeModel.Roles.FilePathRole:
                path = name
                return pathlib.Path(path).name if path else ""
            case FSSpecTreeModel.Roles.FilePermissions:
                return self.model.permissions(name)
            case FSSpecTreeModel.Roles.ProtocolPathRole:
                return self.model.get_protocol_path(name)


class NameColumn(FsSpecColumnItem):
    identifier = "name"
    name = "Name"
    doc = "File name"

    def get_data(self, item: FSSpecTreeModel.TreeItem, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                return pathlib.Path(path).name if (path := item.obj["name"]) else ""
            case constants.DECORATION_ROLE:
                return _icon_provider.get_icon(core.QFileInfo(item.obj["name"]))
            case _:
                return super().get_data(item, role)

    def set_data(self, item, value, role):
        match role:
            case constants.EDIT_ROLE:
                item.set_name(value)


class PathColumn(FsSpecColumnItem):
    identifier = "path"
    name = "Path"
    doc = "File path"

    def get_data(self, item: FSSpecTreeModel.TreeItem, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                return item.obj["name"] or ""
            case _:
                return super().get_data(item, role)


class SizeColumn(FsSpecColumnItem):
    identifier = "size"
    name = "Size"
    doc = "File size."
    precision = 2
    fmt: core._locale.DataSizeFormatStr = "iec"

    def get_data(self, item: FSSpecTreeModel.TreeItem, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                if (size := item.obj["size"]) > 0:
                    return core.Locale().get_formatted_data_size(
                        size, precision=self.precision, fmt=self.fmt
                    )
                else:
                    return ""
            case constants.SORT_ROLE:
                return item.obj["size"]
            case _:
                return super().get_data(item, role)


class TypeColumn(FsSpecColumnItem):
    identifier = "type"
    name = "Type"
    doc = "Type of given path."

    def get_data(self, item: FSSpecTreeModel.TreeItem, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                return item.obj["type"] or ""
            case _:
                return super().get_data(item, role)


class CreatedColumn(FsSpecColumnItem):
    identifier = "created"
    name = "Created"
    doc = "Creation date of file."
    display_format = "%m/%d/%Y, %H:%M:%S"

    def get_data(self, item: FSSpecTreeModel.TreeItem, role: constants.ItemDataRole):
        if not item.obj.get("created"):
            return None
        created = datetime.datetime.fromtimestamp(item.obj["created"])
        match role:
            case constants.DISPLAY_ROLE:
                return created.strftime(self.display_format)
            case constants.USER_ROLE:
                return created
            case _:
                return super().get_data(item, role)


class ModifiedColumn(FsSpecColumnItem):
    identifier = "mtime"
    name = "Modified"
    doc = "Modified"
    display_format = "%m/%d/%Y, %H:%M:%S"

    def get_data(self, item: FSSpecTreeModel.TreeItem, role: constants.ItemDataRole):
        if not item.obj.get("mtime"):
            return None
        mtime = datetime.datetime.fromtimestamp(item.obj["mtime"])
        match role:
            case constants.DISPLAY_ROLE:
                return mtime.strftime(self.display_format)
            case constants.USER_ROLE:
                return mtime
            case _:
                return super().get_data(item, role)


class PermissionsColumn(FsSpecColumnItem):
    identifier = "mode"
    name = "Permissions"
    doc = "File Permissions"

    def get_data(self, item: FSSpecTreeModel.TreeItem, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                if mode := item.obj.get("mode"):
                    return oct(int(mode))[-4:]
            case _:
                return super().get_data(item, role)


class IsLinkColumn(FsSpecColumnItem):
    identifier = "is_link"
    name = "Is link"
    doc = "Whether file is a symbolic link."

    def get_data(self, item: FSSpecTreeModel.TreeItem, role: constants.ItemDataRole):
        match role:
            case constants.CHECKSTATE_ROLE:
                return self.to_checkstate(item.obj.get("islink"))
            case _:
                return super().get_data(item, role)


class ShaColumn(FsSpecColumnItem):
    identifier = "sha"
    name = "SHA"
    doc = "Hash value."

    def get_data(self, item: FSSpecTreeModel.TreeItem, role: constants.ItemDataRole):
        match role:
            case constants.DISPLAY_ROLE:
                return item.obj.get("sha") or ""
            case _:
                return super().get_data(item, role)


FsSpecProtocolStr = str


class FSSpecTreeModel(
    widgets.filesystemmodel.FileSystemModelMixin, itemmodels.ColumnItemModel
):
    """Tree model to display filesystems supported by fsspec.

    The model offers the same API as QFileSystemModel so that it can work
    easily as a drop-in replacement, only difference being that on
    windows, FsSpec does not support a "real" root folder (aka a listing of drives.)
    """

    SUPPORTS = FsSpecProtocolStr | fsspec.AbstractFileSystem
    DEFAULT_COLUMNS = [
        NameColumn,
        SizeColumn,
        TypeColumn,
    ]
    EXTRA_COLUMNS = [
        PathColumn,
        CreatedColumn,
        ModifiedColumn,
        PermissionsColumn,
        IsLinkColumn,
        ShaColumn,
    ]

    directoryLoaded = core.Signal(str)
    fileRenamed = core.Signal(str, str, str)
    rootPathChanged = core.Signal(str)

    class Roles(enum.IntEnum):
        FileIconRole = constants.DECORATION_ROLE
        FilePathRole = constants.USER_ROLE + 1
        FileNameRole = constants.USER_ROLE + 2
        FilePermissions = constants.USER_ROLE + 3
        ProtocolPathRole = constants.USER_ROLE + 4

    class Option(enum.IntEnum):
        DontWatchForChanges = 1
        DontResolveSymlinks = 2
        DontUseCustomDirectoryIcons = 4

    def __init__(
        self,
        protocol: str = "file",
        root: datatypes.PathType = "",
        show_root: bool = False,
        parent: core.QObject | None = None,
        **kwargs,
    ):
        self.set_protocol(protocol, **kwargs)
        obj = self.fs.info(root)
        columns = self.DEFAULT_COLUMNS + self._get_extra_columns_for_protocol(
            self.fs.protocol
        )
        super().__init__(obj, columns=columns, parent=parent, show_root=show_root)

    @classmethod
    def supports(cls, instance) -> bool:
        return isinstance(instance, fsspec.AbstractFileSystem | str)

    def set_protocol(self, protocol: str | fsspec.AbstractFileSystem, **kwargs):
        if isinstance(protocol, str):
            protocol = fsspec.filesystem(protocol, **kwargs)
        self.fs = protocol

    def _get_extra_columns_for_protocol(self, protocol: str) -> list[FsSpecColumnItem]:
        match self.fs.protocol:
            case "github":
                columns = {"size", "sha"}
            case "file":
                columns = {"path", "created", "mtime", "mode", "is_link"}
            case "smb":
                columns = {"uid", "gid", "mtime", "time"}
            case "memory":
                columns = {"created"}
            case _:
                return self.EXTRA_COLUMNS
        return [col for col in self.EXTRA_COLUMNS if col.identifier in columns]

    def _has_children(self, item: FSSpecTreeModel.TreeItem) -> bool:
        return item.obj["type"] == "directory"

    def _fetch_object_children(
        self, item: FSSpecTreeModel.TreeItem
    ) -> list[FSSpecTreeModel.TreeItem]:
        glob = f"{item.obj['name']}/*/" if item.obj["name"] else "*"
        items = [
            FSSpecTreeModel.TreeItem(obj=i, parent=item)
            for i in self.fs.glob(glob, detail=True).values()
        ]
        # not sure if this should be emitted later?
        self.directoryLoaded.emit(item.obj["name"])
        return items

    def get_protocol_path(self, index: core.QModelIndex | str) -> str:
        """Get protocol path for given index."""
        if isinstance(index, str):
            index = self.index(index)
        protocol = self.fs.protocol
        path = index.data(self.Roles.FilePathRole)
        return f"{protocol}://{path}"

    def get_file_content(self, index_or_path: str | core.QModelIndex):
        if isinstance(index_or_path, str):
            index = self.index(index_or_path)
        else:
            index = index_or_path
        protocol_path = self.get_protocol_path(index)
        # info = self.fs.info(protocol_path)
        with self.fs.open(protocol_path) as file:
            return file.read()

    def fileIcon(self, index: core.QModelIndex | str) -> gui.Icon:
        if isinstance(index, str):
            index = self.index(index)
        tree_item = index.internalPointer()
        if tree_item is None:
            return None
        return _icon_provider.get_icon(core.FileInfo(tree_item.obj["name"]))

    def fileInfo(self, index: core.QModelIndex | str):
        if isinstance(index, str):
            index = self.index(index)
        tree_item = index.internalPointer()
        return None if tree_item is None else core.FileInfo(tree_item.obj["name"])

    def fileName(self, index: core.QModelIndex) -> str:
        tree_item = index.internalPointer()
        return "" if tree_item is None else pathlib.Path(tree_item.obj["name"]).name

    def filePath(self, index: core.QModelIndex) -> str:
        tree_item = index.internalPointer()
        return "" if tree_item is None else tree_item.obj["name"]

    def isDir(self, index: core.QModelIndex | str):
        if isinstance(index, str):
            index = self.index(index)
        tree_item = index.internalPointer()
        return False if tree_item is None else tree_item.obj["type"] == "directory"

    def mkdir(self, index: core.QModelIndex | str, name: str):
        if isinstance(index, str):
            index = self.index(index)
        tree_item = index.internalPointer()
        path = tree_item.obj["name"]
        new_folder = pathlib.Path(path) / name
        self.fs.mkdir(str(new_folder))

    def size(self, index: core.QModelIndex | str) -> int:
        if isinstance(index, str):
            index = self.index(index)
        tree_item = index.internalPointer()
        return 0 if tree_item is None else tree_item.obj["size"]

    def lastModified(self, index: core.QModelIndex | str) -> datetime.datetime | None:
        if isinstance(index, str):
            index = self.index(index)
        tree_item = index.internalPointer()
        return (
            None
            if tree_item is None
            else datetime.datetime.fromtimestamp(tree_item.obj["mtime"])
        )

    def permissions(self, index: core.QModelIndex | str) -> core.QFileDevice.Permission:
        if isinstance(index, str):
            index = self.index(index)
        tree_item = index.internalPointer()
        flag = core.QFileDevice.Permission(0)
        if tree_item is None:
            return flag
        val = oct(int(tree_item.obj["mode"]))[-4:]
        for i in core.filedevice.PERMISSIONS.get_list(int(val, 8)):
            flag |= core.filedevice.PERMISSIONS[i]
        return flag

    def remove(self, index: core.QModelIndex | str) -> bool:
        if isinstance(index, str):
            index = self.index(index)
        tree_item = index.internalPointer()
        path = tree_item.obj["name"]
        # TODO: returned bool should indicate whether operation is possible,
        # not whether file exists after trying.
        with self.remove_row(index.row(), index.parent()):
            self.fs.rm_file(path)
        return self.fs.exists(path)

    def rmdir(self, index: core.QModelIndex | str) -> bool:
        if isinstance(index, str):
            index = self.index(index)
        tree_item = index.internalPointer()
        path = tree_item.obj["name"]
        # TODO: returned bool should indicate whether operation is possible,
        # not whether file exists after trying.
        with self.remove_row(index.row(), index.parent()):
            self.fs.rmdir(path)
        return self.fs.exists(path)

    def setRootPath(self, path: os.PathLike | None) -> core.ModelIndex:
        path = os.fspath(path) if path else self.fs.root_marker
        item = self.fs.info(path)
        self.set_root_item(item)
        self.rootPathChanged.emit(path)

    def rootDirectory(self) -> core.Dir:
        return core.Dir(self.rootPath())

    def rootPath(self) -> str:
        return self._root_item["name"]

    def rename(self, index, value: str):
        tree_item = index.internalPointer()
        path = tree_item.obj["name"]
        old_name = pathlib.Path(path).name
        parent_folder = pathlib.Path(path).parent
        target = str(parent_folder / value)
        self.fs.mv(path, target)
        exists = self.fs.exists(target)
        if exists:
            self.fileRenamed.emit(parent_folder, old_name, value)
        return exists

    def type(self, index) -> str:
        tree_item = index.internalPointer()
        return tree_item.obj["type"]

    def mimeTypes(self):
        return ["text/uri-list"]

    def mimeData(self, indexes):
        mime_data = core.MimeData()
        role = self.Roles.FilePathRole
        urls = [core.Url(i) for idx in indexes if (i := idx.data(role))]
        mime_data.setUrls(urls)
        return mime_data

    def supportedDropActions(self):
        DropAction = constants.DropAction
        return DropAction.MoveAction | DropAction.CopyAction | DropAction.LinkAction

    def flags(self, index: core.ModelIndex) -> constants.ItemFlag:
        return super().flags(index) | constants.DROP_ENABLED | constants.DRAG_ENABLED

    # def dragEnterEvent(self, event):
    #     event.accept() if event.mimeData().hasUrls() else super().dragEnterEvent(event)

    # def removeRows(self, row: int, count: int, parent: core.ModelIndex):
    #     end_row = row + count - 1
    #     with self.remove_rows(row, end_row, parent):
    #         for i in range(end_row, row - 1, -1):
    #             index = self.index(i, 0, parent)
    #             path = index.data(self.Roles.FilePathRole)
    #             self.rm_file(path)
    #     return True

    def remove_items(self, offsets: Iterable[int]):
        for offset in sorted(offsets, reverse=True):
            self.removeRow(offset)

    def add_items(
        self,
        items: Iterable,
        position: int | None = None,
        parent: core.QModelIndex | None = None,
    ):
        position = len(self.items) if position is None else position
        items = list(items)
        # with self.insert_rows(position, position + len(items) - 1, parent):
        #     for i in range(len(items)):
        #         self.items.insert(i + position, items[i])
        #         pass
        #     self.items.extend(items)
        return items

    def canDropMimeData(
        self,
        mime_data: core.QMimeData,
        action: constants.DropAction,
        row: int,
        column: int,
        parent_index: core.QModelIndex,
    ) -> bool:
        return column == 0 and mime_data.hasFormat("text/uri-list")

    def dropMimeData(
        self,
        mime_data: core.QMimeData,
        action: constants.DropAction,
        row: int,
        column: int,
        parent_index: core.QModelIndex,
    ):
        if not self.canDropMimeData(mime_data, action, row, column, parent_index):
            return False
        print(mime_data.urls(), action, row, column, parent_index.data())
        urls = [core.Url(i) for i in mime_data.urls()]
        if not urls:
            return False
        match action:
            case constants.DropAction.MoveAction:
                with self.change_layout():
                    for i in sorted(urls, reverse=True):
                        logger.info(
                            "move %s to %s", i, parent_index.data(self.Roles.FilePathRole)
                        )
            case constants.DropAction.CopyAction:
                pass
            case constants.DropAction.LinkAction:
                pass

        column = min(column, 0)
        if row != -1:
            begin_row = row
        begin_row = self.rowCount(parent_index) + 1
        print(begin_row)
        return True

        # if action == Qt.CopyAction:
        #     # Behavior if parent is a text item
        #     # For example, we select a text and do: CTRL+C CTRL+V
        #     if parent.isValid() and not parent.internalPointer().isFolder():
        #         # We insert copy in parent folder, just below
        #         begin_row = parent.row() + 1
        #         parent = parent.parent()

        #     if parent.isValid() and parent.internalPointer().isFolder():
        #         pass
        # if not items:
        #     return False

        # if action == Qt.CopyAction:
        #     pass

        # r = self.insertItems(items, begin_row, parent)
        # return r

        # return True

    # @functools.singledispatchmethod
    def index(
        self,
        path_or_row: str | int,
        column: int = 0,
        index: core.QModelIndex | None = None,
    ) -> core.ModelIndex:
        if isinstance(path_or_row, int):
            return super().index(path_or_row, column, index)
        return self._iter_path(path_or_row, column, index)

    def _iter_path(
        self,
        target: datatypes.PathType,
        column: int = 0,
        parent: core.ModelIndex | None = None,
    ) -> core.ModelIndex:
        parent = parent or core.ModelIndex()
        target = pathlib.Path(target)
        row_count = self.rowCount(parent)
        for i in range(row_count):
            index = self.index(i, column, parent)
            path = self.filePath(index)
            path = pathlib.Path(path)
            if target == path:
                return index
            elif target.is_relative_to(path):
                if self.canFetchMore(index):
                    self.fetchMore(index)
                return self._iter_path(target, column, index)
        return core.ModelIndex()

    def roleNames(self) -> dict[int, core.QByteArray]:
        return {
            259: core.QByteArray(b"filePermissions"),
            1: core.QByteArray(b"fileIcon"),
            4: core.QByteArray(b"statusTip"),
            3: core.QByteArray(b"toolTip"),
            2: core.QByteArray(b"edit"),
            257: core.QByteArray(b"filePath"),
            0: core.QByteArray(b"display"),
            258: core.QByteArray(b"fileName"),
            5: core.QByteArray(b"whatsThis"),
        }

    def setReadOnly(self, val: bool):
        return NotImplemented

    def isReadOnly(self) -> bool:
        return NotImplemented

    def setOption(self, opt, val):
        return NotImplemented

    def options(self):
        return NotImplemented

    def testOption(self, opt):
        return NotImplemented

    def resolveSymlinks(self, val):
        return NotImplemented

    def setResolveSymlinks(self, val):
        return NotImplemented

    def setNameFilterDisables(self, value):
        return NotImplemented

    def nameFilterDisables(self):
        return NotImplemented

    def setNameFilters(self, filters):
        return NotImplemented

    def nameFilters(self):
        return NotImplemented

    def setFilter(self, filter):
        return NotImplemented

    def filter(self):
        return NotImplemented

    readOnly = core.Property(
        bool,
        isReadOnly,
        setReadOnly,
        doc="Used to set model to read-only",
    )


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()

    model = FSSpecTreeModel("github", org="phil65", repo="prettyqt", path="/")
    # model.set_root_path("/")
    tree = widgets.TreeView()
    tree.setRootIsDecorated(True)
    tree.setup_dragdrop_move()
    tree.setAlternatingRowColors(True)
    tree.set_model(model)
    tree.set_selection_behavior("rows")
    tree.setUniformRowHeights(True)
    tree.setAnimated(True)
    tree.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

import os

from typing import Literal

from prettyqt import core, gui
from prettyqt.utils import bidict, datatypes, get_repr, serializemixin


IconEngineHookStr = Literal["available_sizes", "icon_name", "is_null", "scaled_pixmap"]

ICON_ENGINE_HOOK: bidict[IconEngineHookStr, gui.QIconEngine.IconEngineHook] = bidict(
    # available_sizes=gui.QIconEngine.AvailableSizesHook,
    # icon_name=gui.QIconEngine.IconNameHook,
    is_null=gui.QIconEngine.IconEngineHook.IsNullHook,
    scaled_pixmap=gui.QIconEngine.IconEngineHook.ScaledPixmapHook,
)


class IconEngine(serializemixin.SerializeMixin, gui.QIconEngine):
    """Abstract base class for QIcon renderers."""

    def __repr__(self):
        return get_repr(self)

    def __bool__(self):
        return not self.isNull()

    def add_file(
        self,
        path: datatypes.PathType,
        size: datatypes.SizeType,
        mode: gui.icon.ModeStr | gui.QIcon.Mode,
        state: gui.icon.StateStr | gui.QIcon.State,
    ):
        self.addFile(
            os.fspath(path),
            datatypes.to_size(size),
            gui.icon.MODE.get_enum_value(mode),
            gui.icon.STATE.get_enum_value(state),
        )

    def add_pixmap(
        self,
        pixmap: gui.QPixmap,
        mode: gui.icon.ModeStr | gui.QIcon.Mode,
        state: gui.icon.StateStr | gui.QIcon.State,
    ):
        self.addPixmap(
            pixmap,
            gui.icon.MODE.get_enum_value(mode),
            gui.icon.STATE.get_enum_value(state),
        )

    def get_actual_size(
        self,
        size: datatypes.SizeType,
        mode: gui.icon.ModeStr | gui.QIcon.Mode = "normal",
        state: gui.icon.StateStr | gui.QIcon.State = "off",
    ) -> core.Size:
        return core.Size(
            self.actualSize(
                datatypes.to_size(size),
                gui.icon.MODE.get_enum_value(mode),
                gui.icon.STATE.get_enum_value(state),
            )
        )

    def get_available_sizes(
        self,
        mode: gui.icon.ModeStr | gui.QIcon.Mode = "normal",
        state: gui.icon.StateStr | gui.QIcon.State = "off",
    ) -> list[core.Size]:
        return [
            core.Size(i)
            for i in self.availableSizes(
                gui.icon.MODE.get_enum_value(mode), gui.icon.STATE.get_enum_value(state)
            )
        ]

    def get_pixmap(
        self,
        size: datatypes.SizeType,
        mode: gui.icon.ModeStr | gui.QIcon.Mode = "normal",
        state: gui.icon.StateStr | gui.QIcon.State = "off",
        scale: float | None = None,
    ) -> gui.Pixmap:
        if scale is None:
            return gui.Pixmap(
                self.pixmap(
                    datatypes.to_size(size),
                    gui.icon.MODE.get_enum_value(mode),
                    gui.icon.STATE.get_enum_value(state),
                )
            )
        else:
            return gui.Pixmap(
                self.scaledPixmap(
                    datatypes.to_size(size),
                    gui.icon.MODE.get_enum_value(mode),
                    gui.icon.STATE.get_enum_value(state),
                    scale,
                )
            )


if __name__ == "__main__":
    app = gui.app()
    engine = IconEngine()
    print(repr(engine))



from __future__ import annotations

from collections.abc import Iterator
from typing import Any, Literal

from prettyqt import constants, gui, iconprovider
from prettyqt.qt import QtWidgets
from prettyqt.utils import bidict, datatypes, get_repr, listdelegators, serializemixin


mod = QtWidgets.QTreeWidgetItem

ChildIndicatorPolicyStr = Literal["show", "dont_show", "dont_show_when_childless"]

CHILD_INDICATOR_POLICY: bidict[ChildIndicatorPolicyStr, mod.ChildIndicatorPolicy] = (
    bidict(
        show=mod.ChildIndicatorPolicy.ShowIndicator,
        dont_show=mod.ChildIndicatorPolicy.DontShowIndicator,
        dont_show_when_childless=mod.ChildIndicatorPolicy.DontShowIndicatorWhenChildless,
    )
)


class TreeWidgetItem(serializemixin.SerializeMixin, QtWidgets.QTreeWidgetItem):
    def __repr__(self):
        return get_repr(self)

    def serialize_fields(self):
        data = [
            self.data(i, constants.USER_ROLE)  # type: ignore
            for i in range(self.columnCount())
        ]
        return dict(
            text=[self.text(i) for i in range(self.columnCount())],
            tool_tip=[self.toolTip(i) for i in range(self.columnCount())],
            status_tip=[self.statusTip(i) for i in range(self.columnCount())],
            checkstate=[self.get_checkstate(i) for i in range(self.columnCount())],
            icon=[self.get_icon(i) for i in range(self.columnCount())],
            data=data,
        )

    def __iter__(self) -> Iterator[QtWidgets.QTreeWidgetItem]:
        return iter(self.child(i) for i in range(self.childCount()))

    def __len__(self):
        return self.childCount()

    def __getitem__(
        self, index: int | slice
    ) -> (
        QtWidgets.QTreeWidgetItem
        | listdelegators.ListDelegator[QtWidgets.QTreeWidgetItem]
    ):
        match index:
            case int():
                item = self.child(index)
                if item is None:
                    raise KeyError(index)
                return item
            case slice():
                count = self.childCount() if index.stop is None else index.stop
                values = list(range(count)[index])
                ls = [self.child(i) for i in values]
                return listdelegators.ListDelegator(ls)
            case _:
                raise TypeError(index)

    def __delitem__(self, index: int):
        self.takeChild(index)

    def __add__(self, other: QtWidgets.QTreeWidgetItem) -> TreeWidgetItem:
        self.addChild(other)
        return self

    def setChecked(self, column: int, checked: bool):
        self.setCheckState(
            column,
            constants.CheckState.Checked if checked else constants.CheckState.Unchecked,
        )

    def isChecked(self, col: int) -> bool:
        return self.checkState(col) == constants.CheckState.Checked

    def get_children(
        self, recursive: bool = False
    ) -> listdelegators.ListDelegator[QtWidgets.QTreeWidgetItem]:
        """Get children of this item.

        recursive option is written iteratively to also support original QTreeWidgetItems.
        """
        if not recursive:
            return [self.child(i) for i in range(self.childCount())]
        results = []
        nodes = [self]
        while nodes:
            items = []
            for node in nodes:
                results.append(node)
                items.extend(node.child(i) for i in range(node.childCount()))
            nodes = items
        return listdelegators.ListDelegator(results[1:])

    def get_top_level_items(
        self,
    ) -> listdelegators.ListDelegator[QtWidgets.QTreeWidgetItem]:
        items = [self.topLevelItem(i) for i in range(self.topLevelItemCount())]
        return listdelegators.ListDelegator(items)

    def collapse(self, recursive: bool = False):
        if recursive:
            for i in range(self.childCount()):
                self.child(i).collapse(True)
        self.setExpanded(False)

    def expand(self, recursive: bool = False):
        self.setExpanded(True)
        if recursive:
            for i in range(self.childCount()):
                self.child(i).expand(True)

    def set_size_hint(self, hint: datatypes.SizeType, column: int = 0):
        self.setSizeHint(column, datatypes.to_size(hint))

    def sort_children(self, column: int, descending: bool = False):
        order = constants.DESCENDING if descending else constants.ASCENDING
        self.sortChildren(column, order)

    def set_data(
        self, data: Any, role: constants.ItemDataRoleStr | constants.ItemDataRole | int
    ):
        if not isinstance(role, int):
            role = constants.ITEM_DATA_ROLE.get_enum_value(role)
        super().setData(data, role)

    def get_data(
        self, role: constants.ItemDataRoleStr | constants.ItemDataRole | int
    ) -> Any:
        if not isinstance(role, int):
            role = constants.ITEM_DATA_ROLE.get_enum_value(role)
        return super().data(role)

    def set_icon(self, icon: datatypes.IconType, column: int = 0):
        """Set the icon for the action.

        Args:
            icon: icon to use
            column: column
        """
        icon = iconprovider.get_icon(icon)
        self.setIcon(column, icon)

    def get_background(self, column: int = 0) -> gui.Brush:
        return gui.Brush(self.background(column))

    def get_foreground(self, column: int = 0) -> gui.Brush:
        return gui.Brush(self.foreground(column))

    def get_font(self, column: int = 0) -> gui.Font:
        return gui.Font(self.font(column))

    def get_icon(self, column: int = 0) -> gui.Icon | None:
        return None if (icon := self.icon(column)).isNull() else gui.Icon(icon)

    def set_checkstate(
        self, state: constants.CheckStateStr | constants.CheckState, column: int = 0
    ):
        """Set checkstate of the checkbox.

        Args:
            state: checkstate to use
            column: column
        """
        self.setCheckState(column, constants.CHECK_STATE.get_enum_value(state))

    def get_checkstate(self, column: int = 0) -> constants.CheckStateStr:
        """Return checkstate.

        Args:
            column: column

        Returns:
            checkstate
        """
        return constants.CHECK_STATE.inverse[self.checkState(column)]

    def set_child_indicator_policy(
        self, policy: ChildIndicatorPolicyStr | mod.ChildIndicatorPolicy
    ):
        """Set the child indicator policy.

        Args:
            policy: child indicator policy
        """
        self.setChildIndicatorPolicy(CHILD_INDICATOR_POLICY.get_enum_value(policy))

    def get_child_indicator_policy(self) -> ChildIndicatorPolicyStr:
        """Return current child indicator policy.

        Returns:
            child indicator policy
        """
        return CHILD_INDICATOR_POLICY.inverse[self.childIndicatorPolicy()]

    def get_text_alignment(self, column: int) -> constants.AlignmentStr:
        return constants.ALIGNMENTS.inverse[self.textAlignment(column)]


if __name__ == "__main__":
    item = TreeWidgetItem()



from __future__ import annotations

from typing import Literal

from prettyqt import bluetooth, core
from prettyqt.utils import bidict, get_repr


mod = bluetooth.QBluetoothDeviceInfo


CoreConfigurationStr = Literal[
    "none", "base_rate", "base_rate_and_low_energy", "low_energy"
]
CORE_CONFIGURATION: bidict[CoreConfigurationStr, mod.CoreConfiguration] = bidict(
    none=mod.CoreConfiguration.UnknownCoreConfiguration,
    base_rate=mod.CoreConfiguration.BaseRateCoreConfiguration,
    base_rate_and_low_energy=mod.CoreConfiguration.BaseRateAndLowEnergyCoreConfiguration,
    low_energy=mod.CoreConfiguration.LowEnergyCoreConfiguration,
)


FieldStr = Literal[
    "none",
    "rssi",
    "manufacturer_data",
    "service_data",
]
FIELD: bidict[FieldStr, mod.Field] = bidict(
    none=mod.Field(0),
    rssi=mod.Field.RSSI,
    manufacturer_data=mod.Field.ManufacturerData,
    service_data=mod.Field.ServiceData,
)

MajorDeviceClassStr = Literal[
    "miscellaneous",
    "computer",
    "phone",
    "network",
    "audio_video",
    "peripheral",
    "imaging",
    "wearable",
    "toy",
    "health",
    "uncategorized",
]

MAJOR_DEVICE_CLASS: bidict[MajorDeviceClassStr, mod.MajorDeviceClass] = bidict(
    miscellaneous=mod.MajorDeviceClass.MiscellaneousDevice,
    computer=mod.MajorDeviceClass.ComputerDevice,
    phone=mod.MajorDeviceClass.PhoneDevice,
    network=mod.MajorDeviceClass.NetworkDevice,
    audio_video=mod.MajorDeviceClass.AudioVideoDevice,
    peripheral=mod.MajorDeviceClass.PeripheralDevice,
    imaging=mod.MajorDeviceClass.ImagingDevice,
    wearable=mod.MajorDeviceClass.WearableDevice,
    toy=mod.MajorDeviceClass.ToyDevice,
    health=mod.MajorDeviceClass.HealthDevice,
    uncategorized=mod.MajorDeviceClass.UncategorizedDevice,
)

MinorAudioVideoClassStr = Literal[
    "uncategorized",
    "wearable_headset",
    "hands_free",
    "microphone",
    "loudspeaker",
    "headphones",
    "portable_audio",
    "car_audio",
    "set_top_box",
    "hifi_audio",
    "vcr",
    "video_camera",
    "camcorder",
    "video_monitor",
    "video_display_and_loudspeaker",
    "video_conferencing",
    "gaming",
]

MINOR_AUDIO_VIDEO_CLASS: bidict[MinorAudioVideoClassStr, mod.MinorAudioVideoClass] = (
    bidict(
        uncategorized=mod.MinorAudioVideoClass.UncategorizedAudioVideoDevice,
        wearable_headset=mod.MinorAudioVideoClass.WearableHeadsetDevice,
        hands_free=mod.MinorAudioVideoClass.HandsFreeDevice,
        microphone=mod.MinorAudioVideoClass.Microphone,
        loudspeaker=mod.MinorAudioVideoClass.Loudspeaker,
        headphones=mod.MinorAudioVideoClass.Headphones,
        portable_audio=mod.MinorAudioVideoClass.PortableAudioDevice,
        car_audio=mod.MinorAudioVideoClass.CarAudio,
        set_top_box=mod.MinorAudioVideoClass.SetTopBox,
        hifi_audio=mod.MinorAudioVideoClass.HiFiAudioDevice,
        vcr=mod.MinorAudioVideoClass.Vcr,
        video_camera=mod.MinorAudioVideoClass.VideoCamera,
        camcorder=mod.MinorAudioVideoClass.Camcorder,
        video_monitor=mod.MinorAudioVideoClass.VideoMonitor,
        video_display_and_loudspeaker=mod.MinorAudioVideoClass.VideoDisplayAndLoudspeaker,
        video_conferencing=mod.MinorAudioVideoClass.VideoConferencing,
        gaming=mod.MinorAudioVideoClass.GamingDevice,
    )
)


MinorComputerClassStr = Literal[
    "uncategorized",
    "desktop",
    "server",
    "laptop",
    "handheld_clam_shell",
    "handheld_computer",
    "wearable_computer",
]

MINOR_COMPUTER_CLASS: bidict[MinorComputerClassStr, mod.MinorComputerClass] = bidict(
    uncategorized=mod.MinorComputerClass.UncategorizedComputer,
    desktop=mod.MinorComputerClass.DesktopComputer,
    server=mod.MinorComputerClass.ServerComputer,
    laptop=mod.MinorComputerClass.LaptopComputer,
    handheld_clam_shell=mod.MinorComputerClass.HandheldClamShellComputer,
    handheld_computer=mod.MinorComputerClass.HandheldComputer,
    wearable_computer=mod.MinorComputerClass.WearableComputer,
)


MinorHealthClassStr = Literal[
    "uncategorized",
    "pressure_monitor",
    "thermometer",
    "weight_scale",
    "glucose_meter",
    "pulse_oximeter",
    "data_display",
    "step_counter",
]

MINOR_HEALTH_CLASS: bidict[MinorHealthClassStr, mod.MinorHealthClass] = bidict(
    uncategorized=mod.MinorHealthClass.UncategorizedHealthDevice,
    pressure_monitor=mod.MinorHealthClass.HealthBloodPressureMonitor,
    thermometer=mod.MinorHealthClass.HealthThermometer,
    weight_scale=mod.MinorHealthClass.HealthWeightScale,
    glucose_meter=mod.MinorHealthClass.HealthGlucoseMeter,
    pulse_oximeter=mod.MinorHealthClass.HealthPulseOximeter,
    data_display=mod.MinorHealthClass.HealthDataDisplay,
    step_counter=mod.MinorHealthClass.HealthStepCounter,
)

MinorImagingClassStr = Literal[
    "uncategorized",
    "display",
    "camera",
    "scanner",
    "printer",
]

MINOR_IMAGING_CLASS: bidict[MinorImagingClassStr, mod.MinorImagingClass] = bidict(
    uncategorized=mod.MinorImagingClass.UncategorizedImagingDevice,
    display=mod.MinorImagingClass.ImageDisplay,
    camera=mod.MinorImagingClass.ImageCamera,
    scanner=mod.MinorImagingClass.ImageScanner,
    printer=mod.MinorImagingClass.ImagePrinter,
)

MinorNetworkClassStr = Literal[
    "full_service",
    "load_factor_one",
    "load_factor_two",
    "load_factor_three",
    "load_factor_four",
    "load_factor_five",
    "load_factor_six",
    "none",
]

MINOR_NETWORK_CLASS: bidict[MinorNetworkClassStr, mod.MinorNetworkClass] = bidict(
    full_service=mod.MinorNetworkClass.NetworkFullService,
    load_factor_one=mod.MinorNetworkClass.NetworkLoadFactorOne,
    load_factor_two=mod.MinorNetworkClass.NetworkLoadFactorTwo,
    load_factor_three=mod.MinorNetworkClass.NetworkLoadFactorThree,
    load_factor_four=mod.MinorNetworkClass.NetworkLoadFactorFour,
    load_factor_five=mod.MinorNetworkClass.NetworkLoadFactorFive,
    load_factor_six=mod.MinorNetworkClass.NetworkLoadFactorSix,
    none=mod.MinorNetworkClass.NetworkNoService,
)

MinorPeripheralClassStr = Literal[
    "uncategorized",
    "keyboard",
    "pointing_device",
    "keyboard_with_pointing_device",
    "joystick",
    "gamepad",
    "remote_control",
    "sensing_device",
    "digitizer_tablet",
    "card_reader",
]

per = mod.MinorPeripheralClass
MINOR_PERIPHERAL_CLASS: bidict[MinorPeripheralClassStr, per] = bidict(
    uncategorized=per.UncategorizedPeripheral,
    keyboard=per.KeyboardPeripheral,
    pointing_device=per.PointingDevicePeripheral,
    keyboard_with_pointing_device=per.KeyboardWithPointingDevicePeripheral,
    joystick=per.JoystickPeripheral,
    gamepad=per.GamepadPeripheral,
    remote_control=per.RemoteControlPeripheral,
    sensing_device=per.SensingDevicePeripheral,
    digitizer_tablet=per.DigitizerTabletPeripheral,
    card_reader=per.CardReaderPeripheral,
)

MinorPhoneClassStr = Literal[
    "uncategorized",
    "cellular_phone",
    "cordless_phone",
    "smart_phone",
    "wired_modem_or_voice_gateway",
    "common_isdn_access",
]

MINOR_PHONE_CLASS: bidict[MinorPhoneClassStr, mod.MinorPhoneClass] = bidict(
    uncategorized=mod.MinorPhoneClass.UncategorizedPhone,
    cellular_phone=mod.MinorPhoneClass.CellularPhone,
    cordless_phone=mod.MinorPhoneClass.CordlessPhone,
    smart_phone=mod.MinorPhoneClass.SmartPhone,
    wired_modem_or_voice_gateway=mod.MinorPhoneClass.WiredModemOrVoiceGatewayPhone,
    common_isdn_access=mod.MinorPhoneClass.CommonIsdnAccessPhone,
)

MinorToyClassStr = Literal[
    "uncategorized",
    "robot",
    "vehicle",
    "doll",
    "controller",
    "game",
]

MINOR_TOY_CLASS: bidict[MinorToyClassStr, mod.MinorToyClass] = bidict(
    uncategorized=mod.MinorToyClass.UncategorizedToy,
    robot=mod.MinorToyClass.ToyRobot,
    vehicle=mod.MinorToyClass.ToyVehicle,
    doll=mod.MinorToyClass.ToyDoll,
    controller=mod.MinorToyClass.ToyController,
    game=mod.MinorToyClass.ToyGame,
)

MinorWearableClassStr = Literal[
    "uncategorized",
    "robot",
    "vehicle",
    "doll",
    "controller",
    "game",
]

MINOR_WEARABLE_CLASS: bidict[MinorWearableClassStr, mod.MinorWearableClass] = bidict(
    uncategorized=mod.MinorWearableClass.UncategorizedWearableDevice,
    wrist_watch=mod.MinorWearableClass.WearableWristWatch,
    pager=mod.MinorWearableClass.WearablePager,
    jacket=mod.MinorWearableClass.WearableJacket,
    helmet=mod.MinorWearableClass.WearableHelmet,
    glasses=mod.MinorWearableClass.WearableGlasses,
)

ServiceClassStr = Literal[
    "none",
    "positioning",
    "networking",
    "rendering",
    "capturing",
    "object_transfer",
    "audio",
    "telephony",
    "information",
    "all",
]

SERVICE_CLASS: bidict[ServiceClassStr, mod.ServiceClass] = bidict(
    none=mod.ServiceClass.NoService,
    positioning=mod.ServiceClass.PositioningService,
    networking=mod.ServiceClass.NetworkingService,
    rendering=mod.ServiceClass.RenderingService,
    capturing=mod.ServiceClass.CapturingService,
    object_transfer=mod.ServiceClass.ObjectTransferService,
    audio=mod.ServiceClass.AudioService,
    telephony=mod.ServiceClass.TelephonyService,
    information=mod.ServiceClass.InformationService,
    all=mod.ServiceClass.AllServices,
)


class BluetoothDeviceInfo(bluetooth.QBluetoothDeviceInfo):
    """Stores information about the Bluetooth device."""

    def __repr__(self):
        return get_repr(self, self.get_address(), self.name())

    def __bool__(self):
        """Return true if device info is valid."""
        return self.isValid()

    def get_address(self) -> bluetooth.BluetoothAddress:
        """Get bluetooth address."""
        return bluetooth.BluetoothAddress(self.address())

    def get_device_ids(self) -> list[bluetooth.BluetoothUuid]:
        """Get list of device ids."""
        return [bluetooth.BluetoothUuid(i) for i in self.deviceIds()]

    def get_service_ids(self) -> list[bluetooth.BluetoothUuid]:
        """Get list of service ids."""
        return [bluetooth.BluetoothUuid(i) for i in self.serviceIds()]

    def get_service_uuids(self) -> list[bluetooth.BluetoothUuid]:
        """Get list of service UUIDs."""
        return [bluetooth.BluetoothUuid(i) for i in self.serviceUuids()]

    def get_major_device_class(self) -> MajorDeviceClassStr:
        """Get major device class."""
        return MAJOR_DEVICE_CLASS.inverse[self.majorDeviceClass()]

    def get_minor_device_class(self) -> str:
        """Get minor device class."""
        major = self.get_major_device_class()
        match major:
            case "computer":
                flag = mod.MinorComputerClass(self.minorDeviceClass())
                return MINOR_COMPUTER_CLASS.inverse[flag]
            case "phone":
                flag = mod.MinorPhoneClass(self.minorDeviceClass())
                return MINOR_PHONE_CLASS.inverse[flag]
            case "network":
                flag = mod.MinorNetworkClass(self.minorDeviceClass())
                return MINOR_NETWORK_CLASS.inverse[flag]
            case "audio_video":
                flag = mod.MinorAudioVideoClass(self.minorDeviceClass())
                return MINOR_AUDIO_VIDEO_CLASS.inverse[flag]
            case "peripheral":
                flag = mod.MinorPeripheralClass(self.minorDeviceClass())
                return MINOR_PERIPHERAL_CLASS.inverse[flag]
            case "imaging":
                flag = mod.MinorImagingClass(self.minorDeviceClass())
                return MINOR_IMAGING_CLASS.inverse[flag]
            case "wearable":
                flag = mod.MinorWearableClass(self.minorDeviceClass())
                return MINOR_WEARABLE_CLASS.inverse[flag]
            case "toy":
                flag = mod.MinorToyClass(self.minorDeviceClass())
                return MINOR_TOY_CLASS.inverse[flag]
            case "health":
                flag = mod.MinorHealthClass(self.minorDeviceClass())
                return MINOR_HEALTH_CLASS.inverse[flag]
            case "uncategorized":
                return "misc"
            case _:
                raise ValueError(major)

    def get_service_classes(self) -> list[ServiceClassStr]:
        """Get list of service classes."""
        return SERVICE_CLASS.get_list(self.serviceClasses())


if __name__ == "__main__":
    app = core.app()
    info = BluetoothDeviceInfo()



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtGui


class UndoGroupMixin(core.ObjectMixin):
    def __len__(self) -> int:
        return len(self.stacks())

    def __getitem__(self, index: int) -> QtGui.QUndoStack:
        return self.stacks()[index]


class UndoGroup(UndoGroupMixin, QtGui.QUndoGroup):
    pass



from __future__ import annotations

from .scrollbars.annotatedscrollbar import AnnotatedScrollBar
from .scrollbars.previewscrollbar import PreviewScrollBar
from .scrollbars.smoothscrollbar import SmoothScrollBar

from .layouts.flowlayout import FlowLayout
from .layouts.borderlayout import BorderLayout
from .layouts.multilinelayout import MultiLineLayout

from .labels.elidedlabel import ElidedLabel
from .labels.clickablelabel import ClickableLabel
from .labels.iconlabel import IconLabel
from .labels.iconwidget import IconWidget

from .crossfadewidget import CrossFadeWidget
from .faderwidget import FaderWidget
from .imageviewer import ImageViewer
from .editors.listinput import ListInput
from .booldicttoolbutton import BoolDictToolButton
from .optionalwidget import OptionalWidget
from .widgeteditor import WidgetEditor
from .collapsibleframe import CollapsibleFrame
from .orientedtableview import OrientedTableView
from .autoresizingtextedit import AutoResizePlainTextEdit, AutoResizeTextEdit
from .qobjectpropertiestableview import QObjectPropertiesTableView
from .qobjecthierarchytreeview import QObjectHierarchyTreeView
from .logrecordtableview import LogRecordTableView
from .commandpalette import CommandPalette
from .editors.lineedits import IntLineEdit, FloatLineEdit, UrlLineEdit, StringListEdit
from .editors.keycombinationedit import KeyCombinationEdit
from .editors.singlelinetextedit import SingleLineTextEdit
from .editors.brushedit import BrushEdit
from .editors.rectedit import RectEdit, RectFEdit, RegionEdit
from .editors.enumcombobox import EnumComboBox
from .editors.paletteedit import PaletteEdit
from .editors.cursoredit import CursorEdit
from .editors.iconedit import IconEdit
from .editors.localeedit import LocaleEdit
from .editors.pointedit import PointEdit
from .editors.sizeedit import SizeEdit
from .editors.sizepolicyedit import SizePolicyEdit
from .editors.regexlineedit import RegexLineEdit
from .editors.regexinput import RegexInput
from .editors.mappedcheckbox import MappedCheckBox
from .editors.sliceedit import SliceEdit
from .editors.rangeedit import RangeEdit
from .logtextedit import LogTextEdit
from .editors.enumflagwidget import EnumFlagWidget
from .editors.flagselectionwidget import FlagSelectionWidget
from .editors.stringornumberwidget import StringOrNumberWidget
from .iconbrowser import IconBrowser
from .objectbrowser import ObjectBrowser
from .sidebarwidget import SidebarWidget
from .editors.colorcombobox import ColorComboBox
from .multicombobox import MultiComboBox
from .editors.colorchooserbutton import ColorChooserButton
from .editors.filechooserbutton import FileChooserButton
from .editors.fontchooserbutton import FontChooserButton
from .editors.inputandslider import InputAndSlider
from .editors.spanslider import SpanSlider, SpanSliderWidget
from .labeledslider import LabeledSlider
from .waitingspinner import WaitingSpinner
from .numfilterwidget import NumFilterWidget
from .popupinfo import PopupInfo
from .editors.selectionwidget import SelectionWidget
from .codeeditor import CodeEditor
from .astviewer import AstViewer
from .roundprogressbar import RoundProgressBar
from .subsequencecompleter import SubsequenceCompleter
from .scrollareatocwidget import ScrollAreaTocWidget

# from .stareditor import StarEditor, StarRating
from .timeline import Timeline, VideoSample
from .standardiconswidget import StandardIconsWidget

from .itemviews.filetree import FileTree
from .itemviews.filterheader import FilterHeader


__all__ = [
    "AnnotatedScrollBar",
    "PreviewScrollBar",
    "SmoothScrollBar",
    "IntLineEdit",
    "FloatLineEdit",
    "StringListEdit",
    "UrlLineEdit",
    "ImageViewer",
    "CrossFadeWidget",
    "FaderWidget",
    "ElidedLabel",
    "ListInput",
    "BoolDictToolButton",
    "OptionalWidget",
    "MultiLineLayout",
    "WidgetEditor",
    "CollapsibleFrame",
    "CommandPalette",
    "ClickableLabel",
    "OrientedTableView",
    "QObjectPropertiesTableView",
    "QObjectHierarchyTreeView",
    "LogRecordTableView",
    "FilterHeader",
    "KeyCombinationEdit",
    "SingleLineTextEdit",
    "RegexLineEdit",
    "RegexInput",
    "MappedCheckBox",
    "SliceEdit",
    "RangeEdit",
    "LogTextEdit",
    "EnumFlagWidget",
    "FlagSelectionWidget",
    "StringOrNumberWidget",
    "IconLabel",
    "IconBrowser",
    "IconWidget",
    "BrushEdit",
    "RectEdit",
    "RectFEdit",
    "PaletteEdit",
    "CursorEdit",
    "LocaleEdit",
    "IconEdit",
    "PointEdit",
    "SizeEdit",
    "RegionEdit",
    "SizePolicyEdit",
    "FlowLayout",
    "BorderLayout",
    "ObjectBrowser",
    "SidebarWidget",
    "EnumComboBox",
    "ColorComboBox",
    "MultiComboBox",
    "ColorChooserButton",
    "FileChooserButton",
    "FontChooserButton",
    "InputAndSlider",
    "SpanSlider",
    "NumFilterWidget",
    "SpanSliderWidget",
    "LabeledSlider",
    "WaitingSpinner",
    "RoundProgressBar",
    "PopupInfo",
    "SelectionWidget",
    "CodeEditor",
    "Timeline",
    "AutoResizePlainTextEdit",
    "AutoResizeTextEdit",
    # "StarEditor",
    # "StarRating",
    "AstViewer",
    "VideoSample",
    "StandardIconsWidget",
    "SubsequenceCompleter",
    "FileTree",
    "ScrollAreaTocWidget",
]



from __future__ import annotations

from collections.abc import MutableMapping
from typing import Any

from typing_extensions import Self

from prettyqt.qt import QtCore
from prettyqt.utils import datatypes, get_repr


class UrlQuery(QtCore.QUrlQuery, MutableMapping, metaclass=datatypes.QABCMeta):
    """Way to manipulate a key-value pairs in a URL's query."""

    def __repr__(self):
        return get_repr(self, self.toString())

    def __str__(self):
        return self.toString()

    def __contains__(self, key: str):
        return self.hasQueryItem(key)

    def __add__(self, other: dict) -> Self:
        for k, v in other.items():
            self.addQueryItem(k, str(v))
        return self

    def __delitem__(self, item: str):
        self.removeQueryItem(item)

    def __setitem__(self, key, value):
        items = dict(self.queryItems())
        items[key] = value
        items = list(items.items())
        self.setQueryItems(items)

    def __len__(self):
        return len(self.queryItems())

    def __iter__(self):
        return iter(i[0] for i in self.queryItems())

    def __getitem__(self, key: str):
        if not self.hasQueryItem(key):
            raise KeyError(key)
        return self.queryItemValue(key)

    def serialize(self) -> dict[str, Any]:
        return dict(path=self.toString())

    # def add_query_items(self, **items: str):
    #     for k, v in items.items():
    #         self.addQueryitem(k, v)


if __name__ == "__main__":
    query = UrlQuery()
    query["test"] = "hallo"
    query["test2"] = "hallo"
    query.pop("test")
    print(dict(query))



from __future__ import annotations

from collections.abc import Sequence
import enum

from typing import Literal

from prettyqt import constants, core, gui, widgets


BarStyleStr = Literal["donut", "pie", "line", "expand"]

VALUE_TYPE = ["value", "percent", "max"]
ValueTypeStr = Literal["value", "percent", "max"]

VALUE_MAP: dict[str, ValueTypeStr] = {
    r"%v": "value",
    r"%p": "percent",
    r"%m": "max",
}


class RoundProgressBar(widgets.Widget):
    class Position(float, enum.Enum):
        """Start position of progress bar in degrees."""

        Left = 180.0
        Top = 90.0
        Right = 0.0
        Bottom = -90.0

    class BarStyle(enum.IntEnum):
        """Progress bar style."""

        Donut = 1
        Pie = 2
        Line = 3
        Expand = 4

    class ValueType(enum.IntEnum):
        """Value type."""

        value = 1
        percent = 2
        maximum = 3

    def __init__(self, parent: widgets.QWidget | None = None):
        super().__init__(parent)
        self._min_value = 0.0
        self._max_value = 100.0
        self.current_value = 0.0
        self.null_pos = self.Position.Top
        self.bar_style: RoundProgressBar.BarStyle = self.BarStyle.Donut
        self.outline_pen_width = 1.0
        self.data_pen_width = 1.0
        self._rebuild_brush = False
        self.number_format = "%p%"
        self.decimals = 1
        self._update_flags: ValueTypeStr = "percent"
        self.gradient_data: Sequence[gui.QColor] = []

    def minimum(self):
        return self._min_value

    def maximum(self):
        return self._max_value

    # SETTERS -------------------------------------------------------

    def set_null_position(self, position: RoundProgressBar.Position):
        if position != self.null_pos:
            self.null_pos = position
            self._rebuild_brush = True
            self.update()

    def get_null_position(self):
        return self.null_pos

    def set_bar_style(self, style: BarStyleStr | RoundProgressBar.BarStyle):
        if isinstance(style, str):
            BAR_STYLE = dict(
                donut=RoundProgressBar.BarStyle.Donut,
                pie=RoundProgressBar.BarStyle.Pie,
                line=RoundProgressBar.BarStyle.Line,
                expand=RoundProgressBar.BarStyle.Expand,
            )
            style = BAR_STYLE[style]
        if style != self.bar_style:
            self.bar_style = style
            self._rebuild_brush = True
            self.update()

    def set_outline_pen_width(self, width: float):
        if width != self.outline_pen_width:
            self.outline_pen_width = width
            self.update()

    def set_data_pen_width(self, width: float):
        if width != self.data_pen_width:
            self.data_pen_width = width
            self.update()

    def set_data_colors(self, stop_points: list[gui.QColor]):
        if stop_points != self.gradient_data:
            self.gradient_data = stop_points
            self._rebuild_brush = True
            self.update()

    def set_format(self, val: str):
        if val != self.number_format:
            self.number_format = val
            self._value_format_changed()

    def set_decimals(self, count: int):
        if count >= 0 and count != self.decimals:
            self.decimals = count
            self._value_format_changed()

    # SLOTS ---------------------------------------------------------

    @core.Slot(float, float)
    def set_range(self, minval: float, maxval: float):
        self._min_value = min(minval, maxval)
        self._max_value = max(minval, maxval)
        self.current_value = min(
            self._max_value, max(self._min_value, self.current_value)
        )
        self._rebuild_brush = True
        self.update()

    @core.Slot(float)
    def setMinimum(self, val: float):
        self.set_range(val, self._max_value)

    @core.Slot(float)
    def setMaximum(self, val: float):
        self.set_range(self._min_value, val)

    @core.Slot(float)
    def set_value(self, val: float):
        if self.current_value != val:
            self.current_value = min(self._max_value, max(self._min_value, val))
            self.update()

    def get_value(self) -> float:
        return self.current_value

    # PAINTING ------------------------------------------------------

    def paintEvent(self, event: gui.PaintEvent):
        outer_radius = min(self.width(), self.height())
        rect = core.RectF(1, 1, outer_radius - 2, outer_radius - 2)
        with gui.Painter(self) as painter:
            painter.use_antialiasing()
            if self.bar_style != self.BarStyle.Line:
                self._rebuild_data_brush_if_needed()
            painter.fillRect(0, 0, outer_radius, outer_radius, self.palette().window())
            self._draw_base(painter, rect)
            self._draw_value(painter, rect, self.current_value)
            inner_rect, inner_radius = self._calculate_inner_rect(outer_radius)
            self._draw_inner_background(painter, inner_rect)
            self._draw_text(painter, inner_rect, inner_radius, self.current_value)

    def _draw_base(self, painter: gui.Painter, rect: core.RectF):
        match self.bar_style:
            case self.BarStyle.Donut:
                color = self.palette().shadow().color()
                painter.set_pen(color=color, width=self.outline_pen_width)
                painter.setBrush(self.palette().base())
                painter.drawEllipse(rect)
            case self.BarStyle.Line:
                base_color = self.palette().base().color()
                painter.set_pen(color=base_color, width=self.outline_pen_width)
                painter.setBrush(constants.BrushStyle.NoBrush)
                width = self.outline_pen_width / 2
                adjusted = rect.adjusted(width, width, -width, -width)
                painter.drawEllipse(adjusted)
            case self.BarStyle.Pie | self.BarStyle.Expand:
                base_color = self.palette().base().color()
                painter.set_pen(color=base_color, width=self.outline_pen_width)
                painter.setBrush(self.palette().base())
                painter.drawEllipse(rect)

    def _draw_value(self, painter: gui.Painter, rect: core.RectF, value: float):
        if value == self._min_value:
            return
        diff = self.current_value - self._min_value
        value_range = self._max_value - self._min_value
        delta = max(value_range / diff, 0)
        match self.bar_style:
            case self.BarStyle.Expand:
                painter.setBrush(self.palette().highlight())
                color = self.palette().shadow().color()
                painter.set_pen(color=color, width=self.data_pen_width)
                radius = (rect.height() / 2) / delta
                painter.drawEllipse(rect.center(), radius, radius)
            case self.BarStyle.Line:
                color = self.palette().highlight().color()
                painter.set_pen(color=color, width=self.data_pen_width)
                painter.setBrush(constants.BrushStyle.NoBrush)
                pen_width = self.outline_pen_width / 2
                adjusted = rect.adjusted(pen_width, pen_width, -pen_width, -pen_width)
                if value == self._max_value:
                    painter.drawEllipse(adjusted)
                else:
                    arc_length = 360 / delta
                    arc_length = int(-arc_length * 16)
                    painter.drawArc(adjusted, int(self.null_pos * 16), arc_length)
            case self.BarStyle.Donut | self.BarStyle.Pie:
                data_path = gui.PainterPath()
                data_path.set_fill_rule("winding")
                if value == self._max_value:
                    data_path.addEllipse(rect)
                else:
                    arc_length = 360 / delta
                    center_point = rect.center()
                    data_path.moveTo(center_point)
                    data_path.arcTo(rect, self.null_pos, -arc_length)
                    data_path.lineTo(center_point)
                painter.setBrush(self.palette().highlight())
                shadow_color = self.palette().shadow().color()
                painter.set_pen(color=shadow_color, width=self.data_pen_width)
                painter.drawPath(data_path)

    def _calculate_inner_rect(self, outer_radius: float) -> tuple[core.RectF, float]:
        if self.bar_style in (self.BarStyle.Line, self.BarStyle.Expand):
            inner_radius = outer_radius - self.outline_pen_width
        else:
            inner_radius = outer_radius * 0.75
        delta = (outer_radius - inner_radius) / 2
        inner_rect = core.RectF(delta, delta, inner_radius, inner_radius)
        return inner_rect, inner_radius

    def _draw_inner_background(self, painter: gui.Painter, inner_rect: core.RectF):
        if self.bar_style == self.BarStyle.Donut:
            painter.setBrush(self.palette().base())
            painter.drawEllipse(inner_rect)

    def _draw_text(
        self,
        painter: gui.Painter,
        inner_rect: core.RectF,
        inner_radius: float,
        value: float,
    ):
        if not self.number_format:
            return
        font = self.get_font()
        font.setPixelSize(10)
        metrics = gui.FontMetricsF(font)
        max_width = metrics.horizontalAdvance(self._value_to_text(self._max_value))
        delta = inner_radius / max_width
        font_size = int(font.pixelSize() * delta * 0.75)
        font.setPixelSize(max(font_size, 1))
        painter.setFont(font)
        painter.setPen(self.palette().text().color())
        text = self._value_to_text(value)
        painter.drawText(inner_rect, constants.ALIGN_CENTER, text)  # type: ignore

    def _value_to_text(self, value: float) -> str:
        text_to_draw = self.number_format
        match self._update_flags:
            case "value":
                val = round(value, self.decimals)
                return text_to_draw.replace(r"%v", str(val))
            case "percent":
                diff = self._max_value - self._min_value
                pct = (value - self._min_value) / diff * 100
                val = round(pct, self.decimals)
                return text_to_draw.replace(r"%p", str(val))
            case "max":
                val = round(self._max_value - self._min_value + 1, self.decimals)
                return text_to_draw.replace(r"%m", str(val))
            case _:
                raise ValueError(self._update_flags)

    def _value_format_changed(self):
        for k, v in VALUE_MAP.items():
            if k in self.number_format:
                self._update_flags = v
        self.update()

    def _rebuild_data_brush_if_needed(self):
        if not self._rebuild_brush or not self.gradient_data:
            return
        self._rebuild_brush = False
        if self.bar_style == "expand":
            data_brush = gui.RadialGradient(0.5, 0.5, 0.5, 0.5, 0.5)
            for i in self.gradient_data:
                data_brush[i[0]] = i[1]
        else:
            data_brush = gui.ConicalGradient(0.5, 0.5, self.null_pos)
            for i in self.gradient_data:
                data_brush[1 - i[0]] = i[1]
        data_brush.set_coordinate_mode("stretch_to_device")
        with self.edit_palette() as palette:
            palette.set_brush("highlight", data_brush)

    nullPosition = core.Property(
        enum.Enum,
        get_null_position,
        set_null_position,
        doc="Position for value = 0",
    )


if __name__ == "__main__":
    app = widgets.app()
    pb = RoundProgressBar()
    pb.set_bar_style("pie")
    pb.show()
    app.exec()



from __future__ import annotations

from prettyqt import widgets
from prettyqt.qt import QtCore, QtHelp
from prettyqt.utils import datatypes


class HelpContentWidget(widgets.TreeViewMixin, QtHelp.QHelpContentWidget):
    """Tree view for displaying help content model items."""

    def index_of(self, url: datatypes.UrlType) -> QtCore.QModelIndex | None:
        if isinstance(url, str):
            url = QtCore.QUrl(url)
        idx = self.indexOf(url)
        return idx if idx.isValid() else None

    @classmethod
    def setup_example(cls):
        return None



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtHelp


class HelpContentItem:
    """Item for use with QHelpContentModel."""

    def __init__(self, item: QtHelp.QHelpContentItem):
        self.item = item

    def __getattr__(self, val):
        return getattr(self.item, val)

    def __len__(self):
        return self.childCount()

    def get_url(self) -> core.Url:
        return core.Url(self.url())



from __future__ import annotations

from collections.abc import Callable
import sys
import traceback

from typing import Literal

from prettyqt import constants, gui, iconprovider, widgets
from prettyqt.utils import bidict, datatypes


IconStr = Literal["none", "information", "warning", "critical", "question"]

ICONS: bidict[IconStr, widgets.QMessageBox.Icon] = bidict(
    none=widgets.QMessageBox.Icon.NoIcon,
    information=widgets.QMessageBox.Icon.Information,
    warning=widgets.QMessageBox.Icon.Warning,
    critical=widgets.QMessageBox.Icon.Critical,
    question=widgets.QMessageBox.Icon.Question,
)

StandardButtonStr = Literal[
    "none",
    "cancel",
    "ok",
    "save",
    "open",
    "close",
    "discard",
    "apply",
    "reset",
    "restore_defaults",
    "help",
    "save_all",
    "yes",
    "yes_to_all",
    "no",
    "no_to_all",
    "abort",
    "retry",
    "ignore",
]

STANDARD_BUTTON: bidict[StandardButtonStr, widgets.QMessageBox.StandardButton] = bidict(
    none=widgets.QMessageBox.StandardButton.NoButton,
    cancel=widgets.QMessageBox.StandardButton.Cancel,
    ok=widgets.QMessageBox.StandardButton.Ok,
    save=widgets.QMessageBox.StandardButton.Save,
    open=widgets.QMessageBox.StandardButton.Open,
    close=widgets.QMessageBox.StandardButton.Close,
    discard=widgets.QMessageBox.StandardButton.Discard,
    apply=widgets.QMessageBox.StandardButton.Apply,
    reset=widgets.QMessageBox.StandardButton.Reset,
    restore_defaults=widgets.QMessageBox.StandardButton.RestoreDefaults,
    help=widgets.QMessageBox.StandardButton.Help,
    save_all=widgets.QMessageBox.StandardButton.SaveAll,
    yes=widgets.QMessageBox.StandardButton.Yes,
    yes_to_all=widgets.QMessageBox.StandardButton.YesToAll,
    no=widgets.QMessageBox.StandardButton.No,
    no_to_all=widgets.QMessageBox.StandardButton.NoToAll,
    abort=widgets.QMessageBox.StandardButton.Abort,
    retry=widgets.QMessageBox.StandardButton.Retry,
    ignore=widgets.QMessageBox.StandardButton.Ignore,
)

ButtonRoleStr = Literal[
    "invalid",
    "accept",
    "reject",
    "destructive",
    "action",
    "help",
    "yes",
    "no",
    "apply",
    "reset",
]

BUTTON_ROLE: bidict[ButtonRoleStr, widgets.QMessageBox.ButtonRole] = bidict(
    invalid=widgets.QMessageBox.ButtonRole.InvalidRole,
    accept=widgets.QMessageBox.ButtonRole.AcceptRole,
    reject=widgets.QMessageBox.ButtonRole.RejectRole,
    destructive=widgets.QMessageBox.ButtonRole.DestructiveRole,
    action=widgets.QMessageBox.ButtonRole.ActionRole,
    help=widgets.QMessageBox.ButtonRole.HelpRole,
    yes=widgets.QMessageBox.ButtonRole.YesRole,
    no=widgets.QMessageBox.ButtonRole.NoRole,
    apply=widgets.QMessageBox.ButtonRole.ApplyRole,
    reset=widgets.QMessageBox.ButtonRole.ResetRole,
)


class MessageBox(widgets.DialogMixin, widgets.QMessageBox):
    """Modal dialog for informing the user (and for receiving an answer)."""

    def __init__(
        self,
        icon: datatypes.IconType | IconStr = None,
        buttons: list[StandardButtonStr | widgets.QMessageBox.StandardButton]
        | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.set_icon(icon)
        self.setWindowFlags(
            constants.WindowType.Dialog
            | constants.WindowType.WindowTitleHint
            | constants.WindowType.CustomizeWindowHint
        )
        if isinstance(buttons, list):
            for b in buttons:
                self.add_button(b)

    def get_button(
        self, button: widgets.QMessageBox.StandardButton | StandardButtonStr
    ) -> widgets.QAbstractButton:
        return self.button(STANDARD_BUTTON.get_enum_value(button))

    @classmethod
    def message(
        cls,
        text: str,
        title: str = "",
        icon: datatypes.IconType = None,
        detail_text: str | None = None,
    ) -> str:
        m = cls("none", title, text)
        m.set_icon(icon)
        if detail_text is not None:
            m.setDetailedText(detail_text)
        return m.show_blocking()

    @classmethod
    def show_exception(cls, exception: Exception):
        exctype, value = sys.exc_info()[:2]
        tb = traceback.format_exc()
        dlg = cls(text=str(value), title=str(exctype), icon="critical", details=tb)
        dlg.show_blocking()

    def set_icon(self, icon: datatypes.IconType | IconStr):
        if icon in ICONS:
            self.setIcon(ICONS[icon])
        else:
            ico = iconprovider.get_icon(icon)
            self.setIconPixmap(ico.get_pixmap(size=64))

    def show_blocking(self) -> StandardButtonStr:
        return STANDARD_BUTTON.inverse[self.exec()]

    def get_icon_pixmap(self) -> gui.Pixmap | None:
        pix = self.iconPixmap()
        return None if pix.isNull() else gui.Pixmap(pix)

    def set_standard_buttons(
        self, buttons: list[StandardButtonStr | widgets.QMessageBox.StandardButton]
    ):
        flag = self.StandardButton.NoButton
        for val in buttons:
            flag |= STANDARD_BUTTON.get_enum_value(val)
        self.setStandardButtons(flag)

    def get_standard_buttons(self) -> list[StandardButtonStr]:
        return STANDARD_BUTTON.get_list(self.standardButtons())

    def add_button(
        self, button: StandardButtonStr | widgets.QMessageBox.StandardButton
    ) -> widgets.QPushButton:
        """Add a default button.

        Args:
            button: button to add

        Returns:
            created button
        """
        return self.addButton(STANDARD_BUTTON.get_enum_value(button))

    def add_custom_button(
        self,
        button: str,
        role: ButtonRoleStr | widgets.QMessageBox.ButtonRole,
        callback: Callable | None = None,
    ) -> widgets.QPushButton:
        btn = self.addButton(button, BUTTON_ROLE.get_enum_value(role))
        if callback:
            btn.clicked.connect(callback)

    # @classmethod
    # def show_exception(cls, exception):
    #     header = str(exception[0])
    #     error_text = str(exception[1])
    #     widgets.MessageBox.message(error_text, header, "mdi.exclamation")

    def set_text_format(
        self, text_format: constants.TextFormatStr | constants.TextFormat
    ):
        """Set the text format.

        Args:
            text_format: text format to use
        """
        self.setTextFormat(constants.TEXT_FORMAT.get_enum_value(text_format))

    def get_text_format(self) -> constants.TextFormatStr:
        """Return current text format.

        Returns:
            text format
        """
        return constants.TEXT_FORMAT.inverse[self.textFormat()]

    def set_escape_button(self, button: StandardButtonStr | widgets.QAbstractButton):
        if isinstance(button, widgets.QAbstractButton):
            self.setEscapeButton(button)
        else:
            self.setEscapeButton(STANDARD_BUTTON.get_enum_value(button))

    def set_default_button(self, button: StandardButtonStr | widgets.QPushButton):
        if isinstance(button, widgets.QPushButton):
            self.setDefaultButton(button)
        else:
            self.setDefaultButton(STANDARD_BUTTON.get_enum_value(button))


if __name__ == "__main__":
    app = widgets.app()
    ret = MessageBox(icon="warning", window_title="header", text="text")
    ret.set_standard_buttons(["ok", "cancel"])
    ret.add_custom_button("tt", "accept", callback=lambda: print("click"))
    ret.show()
    app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.utils import bidict, get_repr


AccessStr = Literal["private", "protected", "public"]

ACCESS: bidict[AccessStr, core.QMetaMethod.Access] = bidict(
    private=core.QMetaMethod.Access.Private,
    protected=core.QMetaMethod.Access.Protected,
    public=core.QMetaMethod.Access.Public,
)

MethodTypeStr = Literal["method", "signal", "slot", "constructor"]

METHOD_TYPE: bidict[MethodTypeStr, core.QMetaMethod.MethodType] = bidict(
    method=core.QMetaMethod.MethodType.Method,
    signal=core.QMetaMethod.MethodType.Signal,
    slot=core.QMetaMethod.MethodType.Slot,
    constructor=core.QMetaMethod.MethodType.Constructor,
)


class MetaMethod:
    """Meta-data about an member method."""

    def __init__(self, metamethod: core.QMetaMethod):
        self.item = metamethod

    def __getattr__(self, val):
        return getattr(self.item, val)

    def __bool__(self):
        return self.item.isValid()

    def __repr__(self):
        return get_repr(self, self.get_name())

    def get_access(self) -> AccessStr:
        return ACCESS.inverse[self.item.access()]

    def get_method_type(self) -> MethodTypeStr:
        return METHOD_TYPE.inverse[self.item.methodType()]

    def get_method_signature(self) -> str:
        return self.item.methodSignature().data().decode()

    def get_normalized_method_signature(self) -> str:
        """Returns something like 'objectNameChanged(QString)'."""
        sig = self.item.methodSignature()
        normalized = core.QMetaObject.normalizedSignature(sig.data().decode())
        return normalized.data().decode()

    def get_name(self) -> str:
        return self.item.name().data().decode()

    def get_parameters(self) -> list[core.MetaType]:
        count = self.parameterCount()
        return [core.MetaType(self.parameterMetaType(i).id()) for i in range(count)]

    def get_return_type(self) -> core.MetaType:
        return core.MetaType(self.returnMetaType().id())

    def get_parameter_types(self) -> list[str]:
        """Returns sth. like ['QString']."""
        return [i.data().decode() for i in self.parameterTypes()]


if __name__ == "__main__":
    obj = core.Object()
    metaobj = obj.get_metaobject()
    method = metaobj.get_method("objectNameChanged")



from __future__ import annotations

from prettyqt import core, widgets
from prettyqt.utils import get_repr


class GraphicsLineItem(widgets.GraphicsItemMixin, widgets.QGraphicsLineItem):
    """Line item that you can add to a QGraphicsScene."""

    def __repr__(self):
        return get_repr(self, self.get_line())

    def get_line(self) -> core.LineF:
        return core.LineF(self.line())


if __name__ == "__main__":
    item = GraphicsLineItem(core.Point(0, 0), core.Point(2, 2))



from __future__ import annotations

import enum

from prettyqt import core, gui, widgets


class EnumFlagWidget(widgets.ToolButton):
    value_changed = core.Signal(enum.Flag)

    def __init__(self, *args, object_name: str = "enum_flag_widget", **kwargs):
        self._enum_class = None
        self._action_map = {}
        self.button_menu = widgets.Menu(triggered=self._on_menu_click)
        super().__init__(*args, object_name=object_name, **kwargs)
        self.setMenu(self.button_menu)
        self.set_popup_mode("instant")

    def _on_menu_click(self):
        values = []
        value = self.get_value()
        for i in self._enum_class.__members__.values():
            if i.value == 0:
                continue
            # self._action_map[i].setChecked(value & i == i)
            if value & i == i:
                values.append(i)
        text = " | ".join(i.name for i in values) if values else self._enum_class(0).name
        self.set_text(text)
        self.value_changed.emit(value)

    def clear(self):
        self._action_map = {}
        self.button_menu.clear()

    def _set_enum_class(self, enum: enum.EnumMeta | None):
        """Set enum class from which members value should be selected."""
        if enum == self._enum_class:
            return None
        self._enum_class = enum
        self.clear()
        for i in self._enum_class.__members__.values():
            if i.value == 0:
                continue
            action = gui.Action(text=i.name.replace("_", " "), checkable=True)
            action.setData(i)
            self._action_map[i] = action
            self.button_menu.add(action)

    def get_enum_class(self) -> enum.EnumMeta | None:
        """Return current Enum class."""
        return self._enum_class

    def get_value(self) -> enum.Flag:
        """Current value as Enum member."""
        if self._enum_class is None:
            return None
        flag = self._enum_class(0)
        for k, v in self._action_map.items():
            if v.isChecked():
                flag |= k
        return flag

    def set_value(self, value: enum.Flag) -> None:
        """Set value with Enum."""
        if not isinstance(value, enum.Flag):
            value = self._enum_class(value)
        self._set_enum_class(value.__class__)
        if not isinstance(value, self._enum_class):
            raise TypeError(
                "setValue(self, Enum): argument 1 has unexpected type "
                f"{type(value).__name__!r}"
            )
        # this filter shouldnt be needed, see https://bugreports.qt.io/browse/PYSIDE-2369
        values = []
        for i in self._enum_class.__members__.values():
            if i.value == 0:
                continue
            if value & i == i:
                self._action_map[i].setChecked(True)
                values.append(i)
        text = " | ".join(i.name for i in values) if values else self._enum_class(0).name
        self.set_text(text)

    value = core.Property(
        enum.Flag,
        get_value,
        set_value,
        user=True,
        doc="Currently chosen value",
    )
    # enumClass = core.Property(type(enum.Flag), get_enum_class, set_enum_class)


if __name__ == "__main__":
    from prettyqt import constants

    app = widgets.app()
    flag = constants.FocusPolicy
    w = EnumFlagWidget()
    # w.set_value(flag.NoFocus)
    testwidget = widgets.TableView()
    print(type(testwidget.focusPolicy()))
    # print(type(testwidget.focusPolicy()))
    # print(repr(testwidget.focusPolicy()))
    w.set_value(testwidget.focusPolicy())
    w.value_changed.connect(print)
    w.show()
    app.exec()



from __future__ import annotations

import os

from typing import Literal

from prettyqt import core
from prettyqt.qt import QtMultimedia
from prettyqt.utils import bidict, datatypes


EncodingModeStr = Literal[
    "constant_quality", "constant_bit_rate", "average_bit_rate", "two_pass"
]

ENCODING_MODE: bidict[EncodingModeStr, QtMultimedia.QMediaRecorder.EncodingMode] = bidict(
    constant_quality=QtMultimedia.QMediaRecorder.EncodingMode.ConstantQualityEncoding,
    constant_bit_rate=QtMultimedia.QMediaRecorder.EncodingMode.ConstantBitRateEncoding,
    average_bit_rate=QtMultimedia.QMediaRecorder.EncodingMode.AverageBitRateEncoding,
    two_pass=QtMultimedia.QMediaRecorder.EncodingMode.TwoPassEncoding,
)

ErrorStr = Literal["none", "resource", "format", "out_of_space", "location_not_writable"]

ERROR: bidict[ErrorStr, QtMultimedia.QMediaRecorder.Error] = bidict(
    none=QtMultimedia.QMediaRecorder.Error.NoError,
    resource=QtMultimedia.QMediaRecorder.Error.ResourceError,
    format=QtMultimedia.QMediaRecorder.Error.FormatError,
    out_of_space=QtMultimedia.QMediaRecorder.Error.OutOfSpaceError,
    location_not_writable=QtMultimedia.QMediaRecorder.Error.LocationNotWritable,
)

QualityStr = Literal[
    "very_low",
    "low",
    "normal",
    "high",
    "very_high",
]

QUALITY: bidict[QualityStr, QtMultimedia.QMediaRecorder.Quality] = bidict(
    very_low=QtMultimedia.QMediaRecorder.Quality.VeryLowQuality,
    low=QtMultimedia.QMediaRecorder.Quality.LowQuality,
    normal=QtMultimedia.QMediaRecorder.Quality.NormalQuality,
    high=QtMultimedia.QMediaRecorder.Quality.HighQuality,
    very_high=QtMultimedia.QMediaRecorder.Quality.VeryHighQuality,
)

RecorderStateStr = Literal["stopped", "recording", "paused"]

RECORDER_STATE: bidict[RecorderStateStr, QtMultimedia.QMediaRecorder.RecorderState] = (
    bidict(
        stopped=QtMultimedia.QMediaRecorder.RecorderState.StoppedState,
        recording=QtMultimedia.QMediaRecorder.RecorderState.RecordingState,
        paused=QtMultimedia.QMediaRecorder.RecorderState.PausedState,
    )
)


class MediaRecorder(core.ObjectMixin, QtMultimedia.QMediaRecorder):
    def set_output_location(self, url: datatypes.UrlType | os.PathLike | None = None):
        url = datatypes.to_local_url(url)
        self.setOutputLocation(url)

    def get_recorder_state(self) -> RecorderStateStr:
        """Return current recorder state.

        Returns:
            recorder state
        """
        return RECORDER_STATE.inverse[self.recorderState()]

    def get_encoding_mode(self) -> EncodingModeStr:
        """Return current encoding mode.

        Returns:
            encoding mode
        """
        return ENCODING_MODE.inverse[self.encodingMode()]

    def set_encoding_mode(self, mode: EncodingModeStr):
        """Set encoding mode.

        Args:
            mode: encoding mode to use
        """
        self.setEncodingMode(ENCODING_MODE[mode])

    def get_quality(self) -> QualityStr:
        """Return current quality setting.

        Returns:
            quality setting
        """
        return QUALITY.inverse[self.quality()]

    def set_quality(self, quality: QualityStr):
        """Set quality.

        Args:
            quality: quality to use
        """
        self.setQuality(QUALITY[quality])

    def get_error(self) -> ErrorStr:
        """Return error type.

        Returns:
            error type
        """
        return ERROR.inverse[self.error()]


if __name__ == "__main__":
    player = MediaRecorder()
    source = player.get_source()
    print(source.isValid())



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtWidgets


class GraphicsTransformMixin(core.ObjectMixin):
    pass


class GraphicsTransform(GraphicsTransformMixin, QtWidgets.QGraphicsTransform):
    """Abstract base class for building advanced transformations on QGraphicsItems."""


if __name__ == "__main__":
    transform = GraphicsTransform()



"""This module contains the editor widget implementation."""

from __future__ import annotations

import pathlib

from re import Match, Pattern


try:  # pragma: no cover
    import re._constants as sre_constants
except ImportError:  # Python < 3.11
    import sre_constants  # type: ignore

import re

from prettyqt import constants, core, custom_widgets, itemmodels, widgets
from prettyqt.syntaxhighlighters import RegexMatchHighlighter


class RegexEditorWidget(widgets.Widget):
    quick_ref_requested = core.Signal(int)

    def __init__(
        self,
        title: str = "Regex Editor",
        regex: str = "",
        teststring: str = "",
        parent: widgets.QWidget | None = None,
    ):
        super().__init__(window_icon="mdi.regex", window_title=title, parent=parent)
        self.resize(1200, 800)
        self.prog: Pattern | None = None
        self.matches: list[Match] = []
        self.textedit_teststring = widgets.PlainTextEdit(teststring, minimum_width=400)
        self.regexinput = custom_widgets.RegexInput(minimum_width=400)
        self.label_num_matches = widgets.Label("No match", alignment="center")
        self.lineedit_sub = widgets.LineEdit(text_changed=self.update_sub_textedit)
        self.textedit_sub = widgets.PlainTextEdit(read_only=True, minimum_width=400)
        self.table_matches = widgets.TableView()
        self.table_matches.setup_list_style()
        self.textedit_quickref = widgets.TextEdit(
            parent=self,
            read_only=True,
            object_name="textedit_quickref",
            html=(pathlib.Path(__file__).parent / "ref.html").read_text(),
        )
        self.cb_quickref = widgets.CheckBox(
            "Show Regular Expression Quick Reference",
            checked=True,
            state_changed=self.textedit_quickref.setVisible,
        )
        model = itemmodels.RegexMatchesModel()
        self.table_matches.set_model(model)
        self.table_matches.setColumnWidth(0, 60)
        self.table_matches.setColumnWidth(1, 60)
        doc = self.textedit_teststring.document()
        self._highlighter = RegexMatchHighlighter(doc)
        self._highlighter.rehighlight()
        self.regexinput.value_changed.connect(self._update_view)
        self.textedit_teststring.textChanged.connect(self._update_view)
        self.regexinput.pattern = regex
        groupbox = widgets.GroupBox(title="Regular expression")
        layout_toprow = widgets.HBoxLayout()
        layout_toprow.add(self.regexinput)
        grid = widgets.GridLayout(groupbox)
        grid[1, 0] = layout_toprow
        groupbox_teststring = widgets.GroupBox(title="Test strings")
        groupbox_teststring.set_layout("grid")
        groupbox_teststring.box[0, 0] = self.textedit_teststring
        groupbox_sub = widgets.GroupBox(title="Substitution", checkable=True)
        groupbox_sub.toggled.connect(self.lineedit_sub.setVisible)
        groupbox_sub.toggled.connect(self.textedit_sub.setVisible)
        layout_sub = widgets.GridLayout(groupbox_sub)
        layout_sub[0, 0] = self.lineedit_sub
        layout_sub[1, 0] = self.textedit_sub
        layout = self.set_layout("horizontal")
        with layout.get_sub_layout("vertical") as layout:
            layout.add(groupbox)
            layout.add(groupbox_teststring)
            layout.add(groupbox_sub)
            layout.add(self.cb_quickref)
        with layout.get_sub_layout("vertical") as layout:
            layout.add(self.label_num_matches)
            layout.add(self.table_matches)
        layout.add(self.textedit_quickref)
        self._update_view()

    def on_match_list_current_change(self, index_new, index_old):
        model = self.table_matches.model()
        span = model.data(index_new, constants.USER_ROLE)  # type: ignore
        self.textedit_teststring.selecter.select_text(*span)

    def _update_view(self):
        self.prog = None
        self.matches = []
        with self.textedit_teststring.signals_blocked():
            if not self.regexinput.get_pattern():
                self._highlighter.set_spans(None)
                self.table_matches.set_model(None)
                self.label_num_matches.set_text("0 matches")
                return None
            try:
                self.prog = re.compile(
                    self.regexinput.get_pattern(), self.regexinput.get_flags()
                )
            except sre_constants.error:
                self._highlighter.set_spans(None)
                self.table_matches.set_model(None)
                self.label_num_matches.set_text("0 matches")
            else:
                text = self.textedit_teststring.text()
                self.matches = list(self.prog.finditer(text))
                self.label_num_matches.set_text(f"{len(self.matches)} matches")
                model = itemmodels.RegexMatchesModel(self.matches)
                self.table_matches.set_model(model)
                sel_model = self.table_matches.selectionModel()
                sel_model.currentRowChanged.connect(self.on_match_list_current_change)
                spans = [m.span() for m in self.matches]
                self._highlighter.set_spans(spans)
        self.update_sub_textedit()

    def update_sub_textedit(self):
        if self.prog:
            text = self.textedit_teststring.text()
            replace_with = self.lineedit_sub.text()
            substituted = self.prog.sub(replace_with, text)
            self.textedit_sub.set_text(substituted)
        else:
            self.textedit_sub.set_text("")


if __name__ == "__main__":
    app = widgets.app()
    app.setApplicationName("Test")
    app.setOrganizationName("Test")
    teststring = "aa356aa356aa356aa356aa356aa356aa356aa3a356aa356"
    widget = RegexEditorWidget(regex="aa[0-9]", teststring=teststring)
    widget.show()
    app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt.qt import QtCore, QtNetwork
from prettyqt.utils import bidict


NETWORK_ERROR = bidict(
    none=QtNetwork.QNetworkReply.NetworkError.NoError,
    connection_refused=QtNetwork.QNetworkReply.NetworkError.ConnectionRefusedError,
    remote_host_closed=QtNetwork.QNetworkReply.NetworkError.RemoteHostClosedError,
    host_not_found=QtNetwork.QNetworkReply.NetworkError.HostNotFoundError,
    timeout=QtNetwork.QNetworkReply.NetworkError.TimeoutError,
    operaion_canceled=QtNetwork.QNetworkReply.NetworkError.OperationCanceledError,
    ssl_handshake_failed=QtNetwork.QNetworkReply.NetworkError.SslHandshakeFailedError,
    temporary_network_failure=QtNetwork.QNetworkReply.NetworkError.TemporaryNetworkFailureError,
    network_session_failed=QtNetwork.QNetworkReply.NetworkError.NetworkSessionFailedError,
    background_request_not_allowed=QtNetwork.QNetworkReply.NetworkError.BackgroundRequestNotAllowedError,
    too_many_redirects=QtNetwork.QNetworkReply.NetworkError.TooManyRedirectsError,
    insecure_redirect=QtNetwork.QNetworkReply.NetworkError.InsecureRedirectError,
    proxy_connection_refused=QtNetwork.QNetworkReply.NetworkError.ProxyConnectionRefusedError,
    proxy_connection_closed=QtNetwork.QNetworkReply.NetworkError.ProxyConnectionClosedError,
    proxy_not_found=QtNetwork.QNetworkReply.NetworkError.ProxyNotFoundError,
    proxy_timeout=QtNetwork.QNetworkReply.NetworkError.ProxyTimeoutError,
    proxy_authentication_required=QtNetwork.QNetworkReply.NetworkError.ProxyAuthenticationRequiredError,
    content_access_denied=QtNetwork.QNetworkReply.NetworkError.ContentAccessDenied,
    content_operation_not_permitted=QtNetwork.QNetworkReply.NetworkError.ContentOperationNotPermittedError,
    content_not_found=QtNetwork.QNetworkReply.NetworkError.ContentNotFoundError,
    authentication_required=QtNetwork.QNetworkReply.NetworkError.AuthenticationRequiredError,
    content_resend=QtNetwork.QNetworkReply.NetworkError.ContentReSendError,
    content_conflict=QtNetwork.QNetworkReply.NetworkError.ContentConflictError,
    content_gone=QtNetwork.QNetworkReply.NetworkError.ContentGoneError,
    internal_server=QtNetwork.QNetworkReply.NetworkError.InternalServerError,
    operation_not_implemented=QtNetwork.QNetworkReply.NetworkError.OperationNotImplementedError,
    service_unavailable=QtNetwork.QNetworkReply.NetworkError.ServiceUnavailableError,
    protocol_unknown=QtNetwork.QNetworkReply.NetworkError.ProtocolUnknownError,
    protocol_invalid_operation=QtNetwork.QNetworkReply.NetworkError.ProtocolInvalidOperationError,
    unknown_network=QtNetwork.QNetworkReply.NetworkError.UnknownNetworkError,
    unknown_proxy=QtNetwork.QNetworkReply.NetworkError.UnknownProxyError,
    unknown_content=QtNetwork.QNetworkReply.NetworkError.UnknownContentError,
    protocol_failure=QtNetwork.QNetworkReply.NetworkError.ProtocolFailure,
    unknown_server=QtNetwork.QNetworkReply.NetworkError.UnknownServerError,
)

NetworkErrorStr = Literal[
    "none",
    "connection_refused",
    "remote_host_closed",
    "host_not_found",
    "timeout",
    "operaion_canceled",
    "ssl_handshake_failed",
    "temporary_network_failure",
    "network_session_failed",
    "background_request_not_allowed",
    "too_many_redirects",
    "insecure_redirect",
    "proxy_connection_refused",
    "proxy_connection_closed",
    "proxy_not_found",
    "proxy_timeout",
    "proxy_authentication_required",
    "content_access_denied",
    "content_operation_not_permitted",
    "content_not_found",
    "authentication_required",
    "content_resend",
    "content_conflict",
    "content_gone",
    "internal_server",
    "operation_not_implemented",
    "service_unavailable",
    "protocol_unknown",
    "protocol_invalid_operation",
    "unknown_network",
    "unknown_proxy",
    "unknown_content",
    "protocol_failure",
    "unknown_server",
]


class NetworkReply:
    """Contains the data and headers for a request sent with QNetworkAccessManager."""

    def __init__(self, reply: QtNetwork.QNetworkReply):
        self.item = reply

    def __getattr__(self, val):
        return getattr(self.item, val)

    def get_error(self) -> NetworkErrorStr:
        """Return error type.

        Returns:
            error type
        """
        return NETWORK_ERROR.inverse[self.error()]

    def set_raw_headers(self, headers: dict[str, str]):
        for k, v in headers.items():
            self.setRawHeader(
                QtCore.QByteArray(k.encode()), QtCore.QByteArray(v.encode())
            )

    def get_raw_headers(self) -> dict[str, str]:
        return {
            h.data().decode(): self.rawHeader(h).data().decode()
            for h in self.rawHeaderList()
        }



from __future__ import annotations

from collections.abc import Iterator, Sequence
import contextlib
import logging

from typing import Any, Literal, overload

from prettyqt import constants, core
from prettyqt.utils import bidict, helpers, listdelegators, modelhelpers


logger = logging.getLogger(__name__)

CheckIndexOptionStr = Literal[
    "none", "index_is_valid", "do_not_use_parent", "parent_is_invalid"
]

CHECK_INDEX_OPTIONS: bidict[
    CheckIndexOptionStr, core.QAbstractItemModel.CheckIndexOption
] = bidict(
    none=core.QAbstractItemModel.CheckIndexOption.NoOption,
    index_is_valid=core.QAbstractItemModel.CheckIndexOption.IndexIsValid,
    do_not_use_parent=core.QAbstractItemModel.CheckIndexOption.DoNotUseParent,
    parent_is_invalid=core.QAbstractItemModel.CheckIndexOption.ParentIsInvalid,
)

LayoutChangeHintStr = Literal["none", "vertical_sort", "horizontal_sort"]

LAYOUT_CHANGE_HINT: bidict[
    LayoutChangeHintStr, core.QAbstractItemModel.LayoutChangeHint
] = bidict(
    none=core.QAbstractItemModel.LayoutChangeHint.NoLayoutChangeHint,
    vertical_sort=core.QAbstractItemModel.LayoutChangeHint.VerticalSortHint,
    horizontal_sort=core.QAbstractItemModel.LayoutChangeHint.HorizontalSortHint,
)


class AbstractItemModelMixin(core.ObjectMixin):
    IS_RECURSIVE = False  # for docs
    DELEGATE_DEFAULT = None
    DEFAULT_FLAGS = (
        constants.DRAG_ENABLED | constants.IS_ENABLED | constants.IS_SELECTABLE
    )

    def __repr__(self):
        return f"{type(self).__name__}: ({self.rowCount()}, {self.columnCount()})"

    def __len__(self) -> int:
        return self.rowCount()

    def __add__(self, other: core.QAbstractItemModel) -> core.ConcatenateTablesProxyModel:
        proxy = core.ConcatenateTablesProxyModel()
        proxy.addSourceModel(self)
        proxy.addSourceModel(other)
        return proxy

    @overload
    def __getitem__(self, index: tuple[int, int] | int) -> core.ModelIndex:
        ...

    @overload
    def __getitem__(
        self, index: tuple[slice, int] | tuple[int, slice] | tuple[slice, slice]
    ) -> listdelegators.ListDelegator[core.ModelIndex]:
        ...

    def __getitem__(
        self, index: tuple[int | slice, int | slice]
    ) -> core.ModelIndex | listdelegators.ListDelegator[core.ModelIndex]:
        # TODO: do proxies need mapToSource here?
        rowcount = self.rowCount()
        colcount = self.columnCount()
        match index:
            case int() as row, int() as col:
                if row >= rowcount or col >= rowcount:
                    raise IndexError(index)
                return self.index(row, col)
            case (row, col):
                indexes = [
                    self.index(i, j)
                    for i, j in helpers.iter_positions(row, col, rowcount, colcount)
                ]
                return listdelegators.ListDelegator(indexes)
            case int() as row:
                if row >= rowcount:
                    raise IndexError(index)
                # this here breaks PySide6 IPython test...
                return self.index(row, 0)
            case _:
                raise TypeError(index)

    def set_data(
        self,
        index: tuple[int | slice, int | slice] | core.ModelIndex,
        value: Any,
        role=constants.EDIT_ROLE,
    ):
        match index:
            case core.ModelIndex():
                self.setData(index, value, role)
            case (row, col):
                rowcount = self.rowCount()
                colcount = self.columnCount()
                for i, j in helpers.iter_positions(row, col, rowcount, colcount):
                    self.setData(self.index(i, j), value, role)
            case _:
                raise TypeError(index)

    def check_index(
        self,
        index: core.ModelIndex,
        *,
        index_is_valid: bool = False,
        do_not_use_parent: bool = False,
        parent_is_invalid: bool = False,
    ) -> bool:
        flag = core.QAbstractItemModel.CheckIndexOption.NoOption
        if index_is_valid:
            flag |= core.QAbstractItemModel.CheckIndexOption.IndexIsValid
        if do_not_use_parent:
            flag |= core.QAbstractItemModel.CheckIndexOption.DoNotUseParent
        if parent_is_invalid:
            flag |= core.QAbstractItemModel.CheckIndexOption.ParentIsInvalid
        check_flag = core.QAbstractItemModel.CheckIndexOption(0) | flag
        return self.checkIndex(index, check_flag)

    @contextlib.contextmanager
    def change_layout(self):
        """Context manager to change the layout.

        wraps calls with correct signals
        emitted at beginning: layoutAboutToBeChanged
        emitted at end: layoutChanged

        """
        self.layoutAboutToBeChanged.emit()
        yield None
        self.layoutChanged.emit()

    @contextlib.contextmanager
    def reset_model(self):
        """Context manager to reset the model.

        wraps calls with correct signals
        emitted at beginning: beginResetModel
        emitted at end: endResetModel

        """
        self.beginResetModel()
        yield None
        self.endResetModel()

    def get_column_type(
        self,
        column: int,
        *,
        rows_to_check: int = 10,
        role=constants.DISPLAY_ROLE,
    ) -> type | None:
        """Guess column data type by checking values of first rows with given role."""
        to_check = min(rows_to_check, self.rowCount())
        if to_check == 0:
            return None
        # cant combine these or make them a generator, so we do two list comps.
        indexes = [self.index(row, column) for row in range(to_check)]
        values = [self.data(i, role=role) for i in indexes]
        if all(isinstance(i, bool) for i in values):
            return bool
        if all(isinstance(i, str) for i in values):
            return str
        if all(isinstance(i, int) for i in values):
            return int
        if all(isinstance(i, float) for i in values):
            return float
        check_values = [self.data(i, role=constants.CHECKSTATE_ROLE) for i in indexes]
        return bool if None not in check_values else None

    def update_row(self, row: int):
        start_index = self.index(row, 0)
        end_index = self.index(row, self.columnCount() - 1)
        self.dataChanged.emit(start_index, end_index)

    def update_all(self):
        top_left = self.index(0, 0)
        bottom_right = self.index(self.rowCount() - 1, self.columnCount() - 1)
        self.dataChanged.emit(top_left, bottom_right)

    @contextlib.contextmanager
    def remove_row(self, row: int, parent: core.ModelIndex | None = None):
        parent = core.ModelIndex() if parent is None else parent
        self.beginRemoveRows(parent, row, row)
        yield None
        self.endRemoveRows()

    @contextlib.contextmanager
    def remove_rows(
        self,
        first: int | None = None,
        last: int | None = None,
        parent: core.ModelIndex | None = None,
    ):
        parent = core.ModelIndex() if parent is None else parent
        first = first or 0
        last = last if last is not None else self.rowCount()
        self.beginRemoveRows(parent, first, last)
        yield None
        self.endRemoveRows()

    @contextlib.contextmanager
    def remove_columns(
        self,
        first: int | None = None,
        last: int | None = None,
        parent: core.ModelIndex | None = None,
    ):
        parent = core.ModelIndex() if parent is None else parent
        first = first or 0
        last = last if last is not None else self.rowCount()
        self.beginRemoveColumns(parent, first, last)
        yield None
        self.endRemoveColumns()

    @contextlib.contextmanager
    def insert_row(self, row: int, parent: core.ModelIndex | None = None):
        parent = core.ModelIndex() if parent is None else parent
        self.beginInsertRows(parent, row, row)
        yield None
        self.endInsertRows()

    @contextlib.contextmanager
    def insert_rows(
        self,
        first: int | None = None,
        last: int | None = None,
        parent: core.ModelIndex | None = None,
    ):
        parent = core.ModelIndex() if parent is None else parent
        first = first or 0
        last = last if last is not None else self.rowCount()
        self.beginInsertRows(parent, first, last)
        yield None
        self.endInsertRows()

    @contextlib.contextmanager
    def append_rows(self, num_rows: int, parent: core.ModelIndex | None = None):
        parent = core.ModelIndex() if parent is None else parent
        row_count = self.rowCount()
        self.beginInsertRows(parent, row_count, row_count + num_rows - 1)
        yield None
        self.endInsertRows()

    @contextlib.contextmanager
    def insert_columns(
        self,
        first: int | None = None,
        last: int | None = None,
        parent: core.ModelIndex | None = None,
    ):
        parent = core.ModelIndex() if parent is None else parent
        first = first or 0
        last = last if last is not None else self.rowCount()
        self.beginInsertColumns(parent, first, last)
        yield None
        self.endInsertColumns()

    def force_reset(self):
        self.beginResetModel()
        self.endResetModel()

    def force_layoutchange(self):
        self.layoutAboutToBeChanged.emit()
        self.layoutChanged.emit()

    def get_role_names(self) -> bidict[int, str]:
        return bidict({i: v.data().decode() for i, v in self.roleNames().items()})

    def get_breadcrumbs_path(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ) -> list[str]:
        """Get the path for the given index.

        Returns a list containing data of all indexes up to the root for given role.
        """
        pieces = [index.data(role)]
        while (index := index.parent()).isValid():
            pieces.insert(0, index.data(role))
        return pieces

    def iter_tree(
        self,
        parent_index: core.ModelIndex | None = None,
        *,
        depth: int | None = None,
        fetch_more: bool = False,
    ) -> Iterator[core.ModelIndex]:
        """Iter through all indexes of the model tree.

        Arguments:
            parent_index: parent index
            depth: maximum iteration depth
            fetch_more: call fetchMore for all indexes until canFetchMore returns False
        """
        if parent_index is None:
            # TODO: does this always equal AbstractItemView.rootIndex()?
            # parent_index = self.index(0, 0)
            parent_index = core.ModelIndex()
        if parent_index.isValid():
            yield parent_index
            if fetch_more:
                while self.canFetchMore(parent_index):
                    self.fetchMore(parent_index)
        if depth is not None and (depth := depth - 1) < 0:
            return
        for i in range(self.rowCount(parent_index)):
            child_index = self.index(i, 0, parent_index)
            yield from self.iter_tree(child_index, depth=depth, fetch_more=fetch_more)

    def search_tree(
        self,
        value: Any,
        *,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
        parent_index: core.ModelIndex | None = None,
        max_results: int | None = None,
        depth: int | None = None,
    ) -> listdelegators.ListDelegator[core.ModelIndex]:
        """Search the tree for indexes with a given value in given role.

        Compared to QAbstractItemModel.match, this method allows to set a maximum
        search depth and passing several values to search for as a list.

        Arguments:
            value: Item or list of items to search for.
            role: Index role to search in.
            parent_index: start index for searching. If None, whole tree is searched.
            max_results: stop searching after x amount of hits. 'None' means no limit.
            depth: search depth. Search depth. 'None' means no limit.
        """
        results = []
        # This makes it impossible to search for lists. I think thats fine.
        if not isinstance(value, list):
            value = [value]
        for idx in self.iter_tree(parent_index, depth=depth):
            if self.data(idx, role) in value:
                results.append(idx)
                if len(results) == max_results:
                    break
        return listdelegators.ListDelegator(results)

    def get_child_indexes(
        self, index: core.ModelIndex
    ) -> listdelegators.ListDelegator[core.ModelIndex]:
        """Get all child indexes for given index (first column only).

        To get indexes recursively, use iter_tree.

        Arguments:
            index: ModelIndex to get children from
        """
        indexes = [self.index(i, 0, index) for i in range(self.rowCount(index))]
        return listdelegators.ListDelegator(indexes)

    def get_index_key(
        self,
        index: core.ModelIndex,
        *,
        include_column: bool = False,
        parent_index: core.ModelIndex | None = None,
    ) -> tuple[tuple[int, int], ...]:
        """Return a key tuple for given ModelIndex.

        The key tuple consists either of row integers or (row, column) indices
        describing the index position from top to bottom.

        Arguments:
            index: ModelIndex to get a key for
            include_column: whether to include the column in the index key.
            parent_index: Get key up to given ModelIndex. By default, get key up to root.
        """
        key_path = []
        parent = index
        while parent.isValid() and parent != parent_index:
            key = (parent.row(), parent.column()) if include_column else parent.row()
            key_path.append(key)
            parent = parent.parent()
        return tuple(reversed(key_path))

    def index_from_key(
        self,
        key_path: Sequence[tuple[int, int] | int],
        *,
        parent_index: core.ModelIndex | None = None,
    ) -> core.ModelIndex:
        """Return a  ModelIndex for the given key path.

        Arguments:
            key_path: Key path to get an index for.
                      Should be a sequence of either (row, column)  or row indices
            parent_index: ModelIndex to start indexing from. Defaults to root index.
        """
        return modelhelpers.index_from_key(self, key_path, parent_index=parent_index)

    @staticmethod
    def to_checkstate(value: bool):
        return constants.CheckState.Checked if value else constants.CheckState.Unchecked

    def get_header_index_for_value(
        self,
        label,
        orientation: constants.Orientation
        | constants.OrientationStr = constants.HORIZONTAL,
        *,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ) -> int | None:
        o = constants.ORIENTATION.get_enum_value(orientation)
        count = self.columnCount() if o == constants.HORIZONTAL else self.rowCount()
        return next(
            (i for i in range(count) if self.headerData(i, o, role) == label), None
        )

    def get_table_data(
        self,
        *,
        include_index: bool = False,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
        x_range: slice | int | None = None,
        y_range: slice | int | None = None,
        parent_index: core.ModelIndex | None = None,
        use_checkstate_role: bool = False,
    ):
        parent_index = parent_index or core.ModelIndex()
        # if self.canFetchMore(parent_index):
        #     self.fetchMore(parent_index)
        match x_range:
            case None:
                colrange = range(self.columnCount(parent_index))
            case slice():
                stop = x_range.stop or self.columnCount(parent_index)
                colrange = range(x_range.start or 0, stop, x_range.step or 1)
            case int():
                colrange = range(x_range, x_range + 1)
            case _:
                raise TypeError(x_range)

        match y_range:
            case None:
                rowrange = range(self.rowCount(parent_index))
            case slice():
                stop = y_range.stop or self.rowCount(parent_index)
                rowrange = range(y_range.start or 0, stop, y_range.step or 1)
            case int():
                rowrange = range(y_range, y_range + 1)
            case _:
                raise TypeError(y_range)
        data = [
            [self.index(i, j, parent_index).data(role) for j in colrange]
            for i in rowrange
        ]
        if use_checkstate_role:
            check_data, _, __ = self.get_table_data(
                role=constants.CHECKSTATE_ROLE,
                x_range=x_range,
                y_range=y_range,
                parent_index=parent_index,
            )
            for i, row in enumerate(data):
                for j, _column in enumerate(row):
                    if check_data[i][j] is not None:
                        data[i][j] = bool(check_data[i][j])
        h_header = [self.headerData(i, constants.HORIZONTAL) for i in colrange]
        v_header = (
            [self.headerData(i, constants.VERTICAL) for i in rowrange]
            if include_index
            else None
        )
        return data, h_header, v_header

    def to_dataframe(self, **kwargs):
        import pandas as pd

        data, h_header, v_header = self.get_table_data(**kwargs)
        return pd.DataFrame(data=data, columns=h_header, index=v_header)


class AbstractItemModel(AbstractItemModelMixin, core.QAbstractItemModel):
    """The abstract interface for item model classes."""

    # @abc.abstractmethod
    # def index(self, *args, **kwargs):
    #     return NotImplemented

    # # this one is only abstract for a specific signature.
    # # @abc.abstractmethod
    # # def parent(self, *args, **kwargs):
    # #     return NotImplemented

    # @abc.abstractmethod
    # def rowCount(self, *args, **kwargs):
    #     return NotImplemented

    # @abc.abstractmethod
    # def columnCount(self, *args, **kwargs):
    #     return NotImplemented

    # @abc.abstractmethod
    # def data(self, *args, **kwargs):
    #     return NotImp


if __name__ == "__main__":

    class Test(AbstractItemModel):
        pass



from __future__ import annotations

from typing import Literal

from prettyqt.qt import QtGui
from prettyqt.utils import bidict


METRICS = bidict(
    width=QtGui.QPaintDevice.PaintDeviceMetric.PdmWidth,
    height=QtGui.QPaintDevice.PaintDeviceMetric.PdmHeight,
    width_mm=QtGui.QPaintDevice.PaintDeviceMetric.PdmWidthMM,
    height_mm=QtGui.QPaintDevice.PaintDeviceMetric.PdmHeightMM,
    num_colors=QtGui.QPaintDevice.PaintDeviceMetric.PdmNumColors,
    depth=QtGui.QPaintDevice.PaintDeviceMetric.PdmDepth,
    dpi_x=QtGui.QPaintDevice.PaintDeviceMetric.PdmDpiX,
    dpi_y=QtGui.QPaintDevice.PaintDeviceMetric.PdmDpiY,
    physical_dpi_x=QtGui.QPaintDevice.PaintDeviceMetric.PdmPhysicalDpiX,
    physical_dpi_y=QtGui.QPaintDevice.PaintDeviceMetric.PdmPhysicalDpiY,
    pixel_ratio=QtGui.QPaintDevice.PaintDeviceMetric.PdmDevicePixelRatio,
    pixel_ratio_scaled=QtGui.QPaintDevice.PaintDeviceMetric.PdmDevicePixelRatioScaled,
)

MetricStr = Literal[
    "width",
    "height",
    "width_mm",
    "height_mm",
    "num_colors",
    "depth",
    "dpi_x",
    "dpi_y",
    "physical_dpi_x",
    "physical_dpi_y",
    "pixel_ratio",
    "pixel_ratio_scaled",
]


class PaintDeviceMixin:
    def get_metric(self, metric: MetricStr | QtGui.QPaintDevice.PaintDeviceMetric) -> int:
        """Return metric information.

        Args:
            metric: metric information to get

        Returns:
            metric information
        """
        return self.metric(METRICS.get_enum_value(metric))


class PaintDevice(PaintDeviceMixin, QtGui.QPaintDevice):
    """The base class of objects that can be painted on with QPainter."""


if __name__ == "__main__":
    device = PaintDevice()



from __future__ import annotations

from prettyqt import gui, widgets


class WizardPageMixin(widgets.WidgetMixin):
    def set_pixmap(
        self,
        typ: widgets.wizard.WizardPixmapStr | widgets.QWizard.WizardPixmap,
        pixmap: gui.QPixmap | None,
    ):
        if pixmap is None:
            pixmap = gui.QPixmap()
        self.setPixmap(widgets.wizard.WIZARD_PIXMAP.get_enum_value(typ), pixmap)

    def get_pixmap(
        self, typ: widgets.wizard.WizardPixmapStr | widgets.QWizard.WizardPixmap
    ) -> gui.Pixmap | None:
        pix = gui.Pixmap(self.pixmap(widgets.wizard.WIZARD_PIXMAP.get_enum_value(typ)))
        return None if pix.isNull() else pix

    def set_button_text(
        self,
        button_type: widgets.wizard.WizardButtonStr | widgets.QWizard.WizardPixmap,
        value: str,
    ):
        """Set text for given button type.

        Args:
            button_type: button to get text from
            value: text to set

        """
        self.setButtonText(
            widgets.wizard.WIZARD_BUTTON.get_enum_value(button_type), value
        )

    def get_button_text(
        self, button_type: widgets.wizard.WizardButtonStr | widgets.QWizard.WizardButton
    ) -> str:
        """Return text for given button type.

        Args:
            button_type: button to get text from

        Returns:
            Button text
        """
        return self.buttonText(widgets.wizard.WIZARD_BUTTON.get_enum_value(button_type))


class WizardPage(WizardPageMixin, widgets.QWizardPage):
    """The base class for wizard pages."""



from __future__ import annotations

from prettyqt.qt import QtMultimedia


class CameraFormat(QtMultimedia.QCameraFormat):
    def __bool__(self):
        return not self.isNull()


if __name__ == "__main__":
    fmt = CameraFormat()



from __future__ import annotations

from pygments.lexer import RegexLexer, bygroups
from pygments.token import Comment, Keyword, Name, Number, Operator, Text, Token


class CustomLexer(RegexLexer):
    name = "regex"
    aliases = ["regex"]
    filenames: list[str] = []

    tokens = {
        "root": [
            (r"\w+", Name),
            (r"\d+", Number),
            (r"[\s\,\:\-\"\']+", Text),
            (r"[\$\^]", Token),
            (r"[\+\*\.\?]", Operator),
            (
                r"(\()([\?\<\>\!\=\:]{2,3}.+?)(\))",
                bygroups(Keyword.Namespace, Name.Function, Keyword.Namespace),
            ),
            (r"(\()(\?\#.+?)(\))", bygroups(Comment, Comment, Comment)),
            (r"[\(\)]", Keyword.Namespace),
            (r"[\[\]]", Name.Class),
            (r"\\\w", Keyword),
            (r"[\{\}]", Operator),
        ]
    }


# from pygments.token import Other
# from pygments.lexer import include, default

#     tokens = {
#         'root': [
#             (r'\|', Other.Alternate),
#             (r'\(\?[iLmsux]+\)', Other.Directive),
#             (r'\(\?:', Other.Open.NonCapturing),
#             (r'(\(\?P<)(.*?)(>)', Other.Open.NamedCapturing),
#             (r'\(\?=', Other.Open.Lookahead),
#             (r'\(\?!', Other.Open.NegativeLookahead),
#             (r'\(\?<!', Other.Open.NegativeLookbehind),
#             (r'\(\?<', Other.Open.Lookbehind),
#             (r'(\(\?P=)(\w+)(?=\))', Other.Open.ExistsNamed),
#             (r'\(\?\(\d+\)', Other.Open.Exists),
#             (r'\(\?#.*?\)', Other.Comment),
#             (r'\(', Other.Open.Capturing),
#             (r'\)', Other.CloseParen),
#             (r'\[', Other.CharClass, 'charclass_start'),
#             (r'\\[1-9][0-9]?', Other.Backref),
#             include('only_in_verbose'),
#             include('suspicious'),
#             include('meta'),
#             (r'[{}]', Other.UnescapedCurly),  # legal in sre, illegal in regex
#             include('simpleliteral'),
#             (r'[^\\()|\[\]]+', Other.Literals), # TODO
#         ],
#         'suspicious': [
#             # misdone backreferences, tabs, newlines, and bel
#             (r'[\x00-\x08\x0a\x0d]', Other.Suspicious),
#         ],
#         'charclass_start': [
#             (r'\^', Other.NegateCharclass, 'charclass_squarebracket_special'),
#             default('charclass_squarebracket_special'),
#         ],
#         'charclass_squarebracket_special': [
#             (r'\]', Other.Literal.CloseCharClass, 'charclass_rest'),
#             default('charclass_rest'),
#         ],
#         'charclass_rest': [
#             (r'\]', Other.CloseCharClass, '#pop:3'),
#             (r'\\-', Other.EscapedDash),
#             (r'[\-^]', Other.Special),
#             include('simpleliteral'),
#             (r'\\.', Other.Suspicious),
#         ],
#         'meta': [
#             (r'\.', Other.Dot),
#             (r'\^', Other.Anchor.Beginning),
#             (r'\$', Other.Anchor.End),
#             (r'\\b', Other.Anchor.WordBoundary),
#             (r'\\A', Other.Anchor.BeginningOfString),
#             (r'\\Z', Other.Anchor.EndOfString),
#             (r'\*\?', Other.Repetition.NongreedyStar),
#             (r'\*', Other.Repetition.Star),
#             (r'\+\?', Other.Repetition.NongreedyPlus),
#             (r'\+', Other.Repetition.Plus),
#             (r'\?\?', Other.Repetition.NongreedyQuestion),
#             (r'\?', Other.Repetition.Question),
#             (r'\{\d+,(?:\d+)?\}\??', Other.Repetition.Curly),
#             (r'\{,?\d+\}\??', Other.Repetition.Curly),
#         ],
#         'simpleliteral': [
#             (r'[^\\^-]', Other.Literal),
#             (r'\\0[0-7]{0,3}', Other.Literal.Oct),  # \0 is legal
#             (r'\\x[0-9a-fA-F]{2}', Other.Literal.Hex),
#             (r'\\u[0-9a-fA-F]{4}', Other.Literal.Unicode),
#             (r'\\U[0-9a-fA-F]{8}', Other.Literal.LongUnicode),
#             (r'\\[\[\]]', Other.Literal.Bracket),
#             (r'\\[()]', Other.Literal.Paren),
#             (r'\\t', Other.Tab),
#             (r'\\n', Other.Newline),
#             (r'\\\.', Other.Literal.Dot),
#             (r'\\\\', Other.Literal.Backslash),
#             (r'\\\*', Other.Literal.Star),
#             (r'\\\+', Other.Literal.Plus),
#             (r'\\\|', Other.Literal.Alternation),
#             (r'\\\^', Other.Literal.Caret),
#             (r'\\\$', Other.Literal.Dollar),
#             (r'\\\?', Other.Literal.Question),
#             (r'\\[{}]', Other.Literal.Curly),
#             (r'\\\'', Other.Suspicious.Squo),
#             (r'\\\"', Other.Suspicious.Dquo),
#             (r'\\[sSwWdD]', Other.BuiltinCharclass),
#             (r'\\.', Other.Suspicious), # Other unnecessary escapes
#         ],
#         'only_in_verbose': [],
#     }


if __name__ == "__main__":
    from prettyqt import custom_widgets, widgets

    app = widgets.app()
    widget = custom_widgets.RegexLineEdit()
    widget.show()
    widget.value_changed.connect(print)
    app.exec()



from __future__ import annotations

from prettyqt import core, widgets


class SizePolicyEdit(widgets.Widget):
    value_changed = core.Signal(widgets.QSizePolicy)

    def __init__(self, *args, object_name: str = "sizepolicy_edit", **kwargs):
        super().__init__(*args, object_name=object_name, **kwargs)
        self.set_margin(0)
        self.cb_horizontal = widgets.ComboBox()
        self.cb_vertical = widgets.ComboBox()
        self.cb_control_type = widgets.ComboBox()
        layout = self.set_layout("horizontal", margin=0)
        layout.add(widgets.Label("Horizontal:", alignment="center_right"))
        layout.add(self.cb_horizontal)
        layout.add(widgets.Label("Vertical:", alignment="center_right"))
        layout.add(self.cb_vertical)
        layout.add(widgets.Label("Control type:", alignment="center_right"))
        layout.add(self.cb_control_type)
        self.cb_vertical.add_items(widgets.sizepolicy.SIZE_POLICY.keys())
        self.cb_horizontal.add_items(widgets.sizepolicy.SIZE_POLICY.keys())
        self.cb_control_type.add_items(widgets.sizepolicy.CONTROL_TYPE.keys())
        self.cb_horizontal.value_changed.connect(self._on_value_change)
        self.cb_vertical.value_changed.connect(self._on_value_change)
        self.cb_control_type.value_changed.connect(self._on_value_change)

    def _on_value_change(self):
        self._value = self.get_value()
        self.value_changed.emit(self._value)

    def get_value(self) -> core.Point:
        policy = widgets.SizePolicy()
        policy.set_horizontal_policy(self.cb_horizontal.get_value())
        policy.set_vertical_policy(self.cb_vertical.get_value())
        policy.set_control_type(self.cb_control_type.get_value())
        return policy

    def set_value(self, value: widgets.QSizePolicy):
        self._value = value
        self.cb_horizontal.set_value(
            widgets.sizepolicy.SIZE_POLICY.inverse[value.horizontalPolicy()]
        )
        self.cb_vertical.set_value(
            widgets.sizepolicy.SIZE_POLICY.inverse[value.verticalPolicy()]
        )
        self.cb_control_type.set_value(
            widgets.sizepolicy.CONTROL_TYPE.inverse[value.controlType()]
        )

    value = core.Property(
        widgets.QSizePolicy,
        get_value,
        set_value,
        user=True,
        doc="Current value",
    )


if __name__ == "__main__":
    app = widgets.app()
    widget = SizePolicyEdit(window_title="Test")
    widget.set_value(widgets.SizePolicy())
    widget.show()
    app.exec()



from __future__ import annotations

from collections.abc import Generator, Sequence
import functools
import logging

from typing import Any, Literal

from prettyqt import constants, core, widgets
from prettyqt.utils import bidict, classhelpers, datatypes, listdelegators


DelegateStr = Literal[
    "widget",
    "editor",
    "html",
    "markdown",
    "button",
    "no_focus",
    "render_link",
    "star",
    "radio",
    "progress_bar",
]

logger = logging.getLogger(__name__)

EditTriggerStr = Literal[
    "none",
    "current_changed",
    "double_clicked",
    "selected_clicked",
    "edit_key_pressed",
    "any_key_pressed",
    "all",
]

EDIT_TRIGGERS: bidict[EditTriggerStr, widgets.QAbstractItemView.EditTrigger] = bidict(
    none=widgets.QAbstractItemView.EditTrigger.NoEditTriggers,
    current_changed=widgets.QAbstractItemView.EditTrigger.CurrentChanged,
    double_clicked=widgets.QAbstractItemView.EditTrigger.DoubleClicked,
    selected_clicked=widgets.QAbstractItemView.EditTrigger.SelectedClicked,
    edit_key_pressed=widgets.QAbstractItemView.EditTrigger.EditKeyPressed,
    any_key_pressed=widgets.QAbstractItemView.EditTrigger.AnyKeyPressed,
    all=widgets.QAbstractItemView.EditTrigger.AllEditTriggers,
)

SelectionBehaviourStr = Literal["rows", "columns", "items"]

SELECTION_BEHAVIOR: bidict[
    SelectionBehaviourStr, widgets.QAbstractItemView.SelectionBehavior
] = bidict(
    rows=widgets.QAbstractItemView.SelectionBehavior.SelectRows,
    columns=widgets.QAbstractItemView.SelectionBehavior.SelectColumns,
    items=widgets.QAbstractItemView.SelectionBehavior.SelectItems,
)

SelectionModeStr = Literal["single", "extended", "multi", "none"]

SELECTION_MODE: bidict[SelectionModeStr, widgets.QAbstractItemView.SelectionMode] = (
    bidict(
        single=widgets.QAbstractItemView.SelectionMode.SingleSelection,
        extended=widgets.QAbstractItemView.SelectionMode.ExtendedSelection,
        multi=widgets.QAbstractItemView.SelectionMode.MultiSelection,
        none=widgets.QAbstractItemView.SelectionMode.NoSelection,
    )
)

ScrollModeStr = Literal["item", "pixel"]

SCROLL_MODE: bidict[ScrollModeStr, widgets.QAbstractItemView.ScrollMode] = bidict(
    item=widgets.QAbstractItemView.ScrollMode.ScrollPerItem,
    pixel=widgets.QAbstractItemView.ScrollMode.ScrollPerPixel,
)


ScrollHintStr = Literal[
    "ensure_visible", "position_at_top", "position_at_bottom", "position_at_center"
]

SCROLL_HINT: bidict[ScrollHintStr, widgets.QAbstractItemView.ScrollHint] = bidict(
    ensure_visible=widgets.QAbstractItemView.ScrollHint.EnsureVisible,
    position_at_top=widgets.QAbstractItemView.ScrollHint.PositionAtTop,
    position_at_bottom=widgets.QAbstractItemView.ScrollHint.PositionAtBottom,
    position_at_center=widgets.QAbstractItemView.ScrollHint.PositionAtCenter,
)

DragDropModeStr = Literal["none", "drag", "drop", "drag_drop", "internal_move"]

DRAG_DROP_MODE: bidict[DragDropModeStr, widgets.QAbstractItemView.DragDropMode] = bidict(
    none=widgets.QAbstractItemView.DragDropMode.NoDragDrop,
    drag=widgets.QAbstractItemView.DragDropMode.DragOnly,
    drop=widgets.QAbstractItemView.DragDropMode.DropOnly,
    drag_drop=widgets.QAbstractItemView.DragDropMode.DragDrop,
    internal_move=widgets.QAbstractItemView.DragDropMode.InternalMove,
)


DropIndicatorPositionStr = Literal["on_item", "above_item", "below_item", "on_viewport"]

DROP_INDICATOR_POSITION: bidict[
    DropIndicatorPositionStr, widgets.QAbstractItemView.DropIndicatorPosition
] = bidict(
    on_item=widgets.QAbstractItemView.DropIndicatorPosition.OnItem,
    above_item=widgets.QAbstractItemView.DropIndicatorPosition.AboveItem,
    below_item=widgets.QAbstractItemView.DropIndicatorPosition.BelowItem,
    on_viewport=widgets.QAbstractItemView.DropIndicatorPosition.OnViewport,
)


StateStr = Literal[
    "none",
    "dragging",
    "drag_selecting",
    "editing",
    "expanding",
    "collapsing",
    "animating",
]

STATE: bidict[StateStr, widgets.QAbstractItemView.State] = bidict(
    none=widgets.QAbstractItemView.State.NoState,
    dragging=widgets.QAbstractItemView.State.DraggingState,
    drag_selecting=widgets.QAbstractItemView.State.DragSelectingState,
    editing=widgets.QAbstractItemView.State.EditingState,
    expanding=widgets.QAbstractItemView.State.ExpandingState,
    collapsing=widgets.QAbstractItemView.State.CollapsingState,
    animating=widgets.QAbstractItemView.State.AnimatingState,
)


class AbstractItemViewMixin(widgets.AbstractScrollAreaMixin):
    model_changed = core.Signal(core.QAbstractItemModel)

    def __init__(
        self,
        *args,
        horizontal_scroll_mode="pixel",
        vertical_scroll_mode="pixel",
        **kwargs,
    ):
        super().__init__(
            *args,
            horizontal_scroll_mode=horizontal_scroll_mode,
            vertical_scroll_mode=vertical_scroll_mode,
            **kwargs,
        )
        from prettyqt.utils import proxifier

        self.proxifier = proxifier.Proxifier(self)

    def __len__(self) -> int:
        return model.rowCount() if (model := self.model()) is not None else 0

    def _get_map(self):
        maps = super()._get_map()
        maps |= {
            "dragDropMode": DRAG_DROP_MODE,
            "horizontalScrollMode": SCROLL_MODE,
            "verticalScrollMode": SCROLL_MODE,
            "selectionMode": SELECTION_MODE,
            "selectionBehavior": SELECTION_BEHAVIOR,
            "defaultDropAction": constants.DROP_ACTION,
            "textElideMode": constants.TEXT_ELIDE_MODE,
            "editTriggers": EDIT_TRIGGERS,
        }
        return maps

    def selectAll(self):
        """Override, we dont want to selectAll for too many items bc of performance."""
        if self.model() is None:
            return
        if self.model().rowCount() * self.model().columnCount() > 1_000_000:
            logger.warning("Too many cells to select.")
            return
        super().selectAll()

    def set_model(self, model: core.QAbstractItemModel | None):
        """Set the model of this View."""
        # Delete old selection model explicitely, seems to help with memory usage.
        old_model = self.model()
        old_sel_model = self.selectionModel()
        if old_model is not None or model is not None:
            self.setModel(model)
            if model is not None:
                # always set a parent. PySide6 needs one for proxying.
                model.setParent(self)
                self.setSelectionModel(core.ItemSelectionModel(model))
            self.model_changed.emit(model)
        # if old_model:
        #     old_model.deleteLater()
        #     del old_model
        if old_sel_model:
            old_sel_model.deleteLater()
            del old_sel_model
        return model

    def set_model_for(self, data: Any):
        """Set model for given data type.

        Pass any data structure and an appropriate model will be chosen automatically.

        Args:
            data: data to choose model for.
        """
        # we import to collect the models
        from prettyqt import itemmodels  # noqa: F401

        for Klass in classhelpers.get_subclasses(core.QAbstractItemModel):
            if (
                "supports" in Klass.__dict__
                and callable(Klass.supports)
                and Klass.supports(data)
                and Klass.__name__ != "PythonObjectTreeModel"
            ):
                logger.debug(f"found model for data structure {data!r}")
                break
        else:
            raise TypeError("No suiting model found.")
        model = Klass(data, parent=self)
        self.set_model(model)

    def get_model(self, skip_proxies: bool = False) -> core.QAbstractItemModel:
        """Get current model of the ItemView.

        Arguments:
            skip_proxies: Whether to get current model or the non-proxy sourceModel.
        """
        model = self.model()
        if skip_proxies:
            while isinstance(model, core.QAbstractProxyModel):
                model = model.sourceModel()
        return model

    def get_models(
        self, proxies_only: bool = False
    ) -> listdelegators.ListDelegator[core.QAbstractProxyModel]:
        """Get a list of all (proxy) models connected to this view.

        Arguments:
            proxies_only: whether the non-proxy sourceModel should be included.
        """
        model = self.model()
        models = []
        while isinstance(model, core.QAbstractProxyModel):
            models.append(model)
            model = model.sourceModel()
        if (not proxies_only) and model is not None:
            models.append(model)
        return listdelegators.ListDelegator(models)

    def set_current_index(
        self,
        index: core.QModelIndex | tuple | None,
        operation: Literal["select", "deselect", "toggle"] = "select",
        *,
        clear: bool = True,
        current: bool = False,
        expand: Literal["rows", "columns"] | None = None,
    ):
        """Set current index.

        Arguments:
            index: Index to set.
            operation: Whether to select, deselect or toggle the current state.
            clear: Clear Whether to clear previously selected indexes.
            current: Current selection will be updated.
            expand: Whether to expand selection to whole column / row.
        """
        match index:
            case None:
                self.selectionModel().setCurrentIndex(
                    index, core.ItemSelectionModel.SelectionFlag.Clear
                )
                return
            case tuple():
                index = self.model().index(*index)
            case core.QModelIndex():
                pass
            case _:
                raise ValueError(index)
        match operation:
            case "select":
                flag = core.ItemSelectionModel.SelectionFlag.Select
            case "deselect":
                flag = core.ItemSelectionModel.SelectionFlag.Deselect
            case "toggle":
                flag = core.ItemSelectionModel.SelectionFlag.Toggle
            case _:
                raise ValueError(operation)
        if clear:
            flag |= core.ItemSelectionModel.SelectionFlag.Clear
        if current:
            flag |= core.ItemSelectionModel.SelectionFlag.Current
        match expand:
            case "rows":
                flag |= core.ItemSelectionModel.SelectionFlag.Rows
            case "columns":
                flag |= core.ItemSelectionModel.SelectionFlag.Columns
            case None:
                pass
            case _:
                raise ValueError(expand)
        self.selectionModel().setCurrentIndex(index, flag)

    def select_index(
        self,
        index: core.QModelIndex | tuple | None,
        operation: Literal["select", "deselect", "toggle"] = "select",
        *,
        clear: bool = True,
        current: bool = False,
        expand: Literal["rows", "columns"] | None = None,
    ):
        # index = self.model().index(self._selected_index)
        match index:
            case None:
                self.selectionModel().setCurrentIndex(
                    core.ModelIndex(), core.ItemSelectionModel.SelectionFlag.Clear
                )
                return
            case tuple():
                index = self.model().index(*index)
            case core.QModelIndex():
                pass
            case _:
                raise ValueError(index)
        match operation:
            case "select":
                flag = core.ItemSelectionModel.SelectionFlag.Select
            case "deselect":
                flag = core.ItemSelectionModel.SelectionFlag.Deselect
            case "toggle":
                flag = core.ItemSelectionModel.SelectionFlag.Toggle
            case _:
                raise ValueError(operation)
        if clear:
            flag |= core.ItemSelectionModel.SelectionFlag.Clear
        if current:
            flag |= core.ItemSelectionModel.SelectionFlag.Current
        match expand:
            case "rows":
                flag |= core.ItemSelectionModel.SelectionFlag.Rows
            case "columns":
                flag |= core.ItemSelectionModel.SelectionFlag.Columns
            case None:
                pass
            case _:
                raise ValueError(expand)
        self.selectionModel().select(index, flag)

    # def move_selection(self, row_offset: int = 0, column_offset: int = 0):
    #     """Move selection by an offset."""
    #     for index in self.selectedIndexes():
    #         new_idx = self.model().index(
    #             index.row() + row_offset, index.column() + column_offset, index.parent()
    #         )
    #         if new_idx.isValid():
    #             self.set_current_index(new_idx, current=True, expand="rows")

    def move_row_selection(self, dy: int):
        for row in self.selected_rows():
            new_idx = self.model().index(row + dy, 0)
            if new_idx.isValid():
                self.set_current_index(new_idx, current=True, expand="rows")

    def set_delegate(
        self,
        delegate: widgets.QAbstractItemDelegate | DelegateStr | None,
        *,
        column: int | Sequence | None = None,
        row: int | Sequence | None = None,
        persistent: bool = False,
        **kwargs: Any,
    ):
        """Set a item delegate for the view.

        Arguments:
            delegate: Delegate to set. Can also be the id of the delegate.
            column: Column the delegate should be set for.
            row: Row the column should be set for.
            persistent: If True, open persistent editors for given area.
            kwargs: Keyword args to pass to the Delegate ctor if delegate is set by id.
        """
        match delegate:
            case widgets.QAbstractItemDelegate():
                dlg = delegate
            # case "editor":
            #     delegate = itemdelegates.EditorDelegate(parent=self, **kwargs)
            # case "widget":
            #     delegate = itemdelegates.WidgetDelegate(parent=self, **kwargs)
            # case "html":
            #     delegate = itemdelegates.HtmlItemDelegate(parent=self, **kwargs)
            # case "button":
            #     delegate = itemdelegates.ButtonDelegate(parent=self, **kwargs)
            case str():
                Klass = classhelpers.get_class_for_id(
                    widgets.StyledItemDelegate, delegate
                )
                dlg = Klass(parent=self, **kwargs)
            case None:
                dlg = widgets.StyledItemDelegate()
            case _:
                raise ValueError(delegate)
        match column, row:
            case int(), int():
                raise ValueError("Only set column or row, not both.")
            case Sequence(), None:
                for i in column:
                    self.set_delegate(delegate, column=i, row=row, persistent=persistent)
            case int(), None:
                self.setItemDelegateForColumn(column, dlg)
                if persistent:
                    model = self.model()
                    for i in range(model.rowCount()):
                        index = model.index(i, column)
                        self.openPersistentEditor(index)
            case None, Sequence():
                for i in row:
                    self.set_delegate(
                        delegate, column=column, row=i, persistent=persistent
                    )
            case None, int():
                self.setItemDelegateForRow(row, dlg)
                if persistent:
                    model = self.model()
                    for i in range(model.columnCount()):
                        self.openPersistentEditor(model.index(row, i))
            case None, None:
                self.setItemDelegate(dlg)
                if persistent:
                    model = self.model()
                    for i in range(model.rowCount()):
                        for j in range(model.columnCount()):
                            self.openPersistentEditor(model.index(i, j))
        return dlg

    def toggle_select_all(self):
        """Select all items from list (deselect when all selected)."""
        if self.selectionModel() is None:
            return
        if self.selectionModel().hasSelection():
            self.clearSelection()
        else:
            self.selectAll()

    def set_table_color(self, color: str):
        with self.edit_stylesheet() as ss:
            ss.QHeaderView.section.backgroundColor.setValue(color)

    def current_index(self) -> core.QModelIndex | None:
        if (model := self.selectionModel()) is not None:
            return model.currentIndex()
        return None

    def current_data(self, role=constants.USER_ROLE):
        if (model := self.selectionModel()) is not None:
            idx = model.currentIndex()
            return idx.data(role)

    def current_row(self) -> int | None:
        if (model := self.selectionModel()) is not None:
            return model.currentIndex().row()
        return None

    def current_column(self) -> int | None:
        if (model := self.selectionModel()) is not None:
            return model.currentIndex().column()
        return None

    def selected_indexes(self) -> listdelegators.ListDelegator[core.QModelIndex]:
        """Return list of selected indexes in first row."""
        indexes = (x for x in self.selectedIndexes() if x.column() == 0)
        indexes = sorted(indexes, key=lambda x: x.row())
        return listdelegators.ListDelegator(indexes)

    def selected_names(self) -> Generator[Any, None, None]:
        """Return generator yielding item names."""
        return (x.data(constants.NAME_ROLE) for x in self.selected_indexes())

    def selected_rows(self) -> Generator[int, None, None]:
        """Return generator yielding row nums."""
        return (x.row() for x in self.selected_indexes())

    def selected_data(
        self, role: constants.ItemDataRole = constants.USER_ROLE
    ) -> Generator[Any, None, None]:
        """Return generator yielding selected userData.

        Arguments:
            role: ItemRole to return data for.
        """
        return (x.data(role) for x in self.selected_indexes())

    def setup_dragdrop_move(self):
        self.setDragEnabled(True)
        self.setAcceptDrops(True)
        self.setDragDropMode(self.DragDropMode.DragDrop)
        self.setDefaultDropAction(constants.MOVE_ACTION)
        self.setDropIndicatorShown(True)

    def set_edit_triggers(self, *triggers: EditTriggerStr | None):
        items = ["none" if t is None else t for t in triggers]
        flags = EDIT_TRIGGERS.merge_flags(items)
        self.setEditTriggers(flags)

    def get_edit_triggers(self) -> list[EditTriggerStr]:
        return EDIT_TRIGGERS.get_list(self.editTriggers())

    def set_selection_behavior(
        self,
        behavior: SelectionBehaviourStr | widgets.QAbstractItemView.SelectionBehavior,
    ):
        """Set selection behavior for given item view.

        Arguments:
            behavior: Selection behavior
        """
        self.setSelectionBehavior(SELECTION_BEHAVIOR.get_enum_value(behavior))

    def get_selection_behavior(self) -> SelectionBehaviourStr:
        """Return current selection behavior.

        Returns:
            selection behavior
        """
        return SELECTION_BEHAVIOR.inverse[self.selectionBehavior()]

    def get_drop_indicator_position(self) -> DropIndicatorPositionStr:
        """Return position of the drop indicator in relation to the closest item."""
        return DROP_INDICATOR_POSITION.inverse[self.dropIndicatorPosition()]

    def set_drag_drop_mode(
        self, mode: DragDropModeStr | widgets.QAbstractItemView.DragDropMode
    ):
        """Set drag-drop mode for given item view.

        Args:
            mode: drag-drop mode to use
        """
        self.setDragDropMode(DRAG_DROP_MODE.get_enum_value(mode))

    def get_drag_drop_mode(self) -> DragDropModeStr:
        """Return current drag-drop mode."""
        return DRAG_DROP_MODE.inverse[self.dragDropMode()]

    def set_state(self, state: StateStr | widgets.QAbstractItemView.State):
        """Set state for given item view.

        Args:
            state: state to use
        """
        self.setState(STATE.get_enum_value(state))

    def get_state(self) -> StateStr:
        """Return current state."""
        return STATE.inverse[self.state()]

    def set_selection_mode(
        self, mode: SelectionModeStr | widgets.QAbstractItemView.SelectionMode | None
    ):
        """Set selection mode for given item view.

        Args:
            mode: selection mode to use
        """
        if mode is None:
            mode = "none"
        self.setSelectionMode(SELECTION_MODE.get_enum_value(mode))

    def get_selection_mode(self) -> SelectionModeStr:
        """Return current selection mode."""
        return SELECTION_MODE.inverse[self.selectionMode()]

    def set_scroll_mode(self, mode: ScrollModeStr | widgets.QAbstractItemView.ScrollMode):
        """Set the scroll mode for both directions.

        Args:
            mode: mode to set
        """
        self.setHorizontalScrollMode(SCROLL_MODE.get_enum_value(mode))
        self.setVerticalScrollMode(SCROLL_MODE.get_enum_value(mode))

    def set_horizontal_scroll_mode(
        self, mode: ScrollModeStr | widgets.QAbstractItemView.ScrollMode
    ):
        """Set the horizontal scroll mode.

        Args:
            mode: mode to set
        """
        self.setHorizontalScrollMode(SCROLL_MODE.get_enum_value(mode))

    def get_horizontal_scroll_mode(self) -> ScrollModeStr:
        """Return current horizontal scroll mode."""
        return SCROLL_MODE.inverse[self.horizontalScrollMode()]

    def set_vertical_scroll_mode(
        self, mode: ScrollModeStr | widgets.QAbstractItemView.ScrollMode
    ):
        """Set the vertical scroll mode.

        Args:
            mode: mode to set
        """
        self.setVerticalScrollMode(SCROLL_MODE.get_enum_value(mode))

    def get_vertical_scroll_mode(self) -> ScrollModeStr:
        """Return current vertical scroll mode."""
        return SCROLL_MODE.inverse[self.verticalScrollMode()]

    def num_selected(self) -> int:
        """Return amount of selected rows.

        Returns:
            amount of selected rows
        """
        if (model := self.selectionModel()) is not None:
            return len(model.selectedRows())
        return 0

    def jump_to_column(self, col_num: int):
        """Make sure column at given index is visible.

        scrolls to column at given index

        Args:
            col_num: column to scroll to
        """
        if (model := self.model()) is not None:
            idx = model.index(0, col_num)
            self.scrollTo(idx)

    def scroll_to_top(self):
        """Override to use abstractitemview-way of scrolling to top."""
        self.scrollToTop()

    def scroll_to_bottom(self):
        """Override to use abstractitemview-way of scrolling to bottom."""
        self.scrollToBottom()

    def select_last_row(self):
        idx = self.model().createIndex(self.model().rowCount() - 1, 0)
        self.setCurrentIndex(idx)

    def select_first_row(self):
        idx = self.model().index(0, 0)
        self.set_current_index(idx, current=True, expand="rows")

    def scroll_to(
        self,
        index: core.QModelIndex,
        mode: ScrollHintStr | widgets.QAbstractItemView.ScrollHint = "ensure_visible",
    ):
        self.scrollTo(index, SCROLL_HINT.get_enum_value(mode))

    def highlight_when_inactive(self):
        """Highlight items when widget does not have focus."""
        p = self.get_palette()
        p.highlight_inactive()
        self.setPalette(p)

    def set_icon_size(self, size: datatypes.SizeType):
        self.setIconSize(datatypes.to_size(size))

    def get_size_hint_for_column(self, col: int, row_limit: int = 25) -> int:
        """Get a size hint for given column by finding widest cell.

        Arguments:
            col: columnt to get size hint for.
            row_limit: number of rows to check.
        """
        to_check = min(row_limit, self.model().rowCount())
        return max(
            self.sizeHintForIndex(self.model().index(row, col)).width()
            for row in range(to_check)
        )

    def sync_with(
        self,
        table_to_sync: widgets.QAbstractItemView,
        orientation: constants.OrientationStr | constants.Orientation,
    ) -> list[core.QMetaObject.Connection]:
        orientation = constants.ORIENTATION.get_enum_value(orientation)
        """Sync ItemView cell widths and scrollbars with another ItemView.

        Arguments:
            table_to_sync: Table to sync with
            orientation: Whether to sync horizontal or vertical orientation.
        """

        def _table_resized(col, _, new_size, table, orientation):
            if orientation == constants.HORIZONTAL:
                table.setColumnWidth(col, new_size)
            else:
                table.setRowHeight(col, new_size)

        _table_1_resized = functools.partial(
            _table_resized, table=self, orientation=orientation
        )
        _table_2_resized = functools.partial(
            _table_resized, table=table_to_sync, orientation=orientation
        )
        if orientation == constants.VERTICAL:
            h1 = self.v_scrollbar.valueChanged.connect(table_to_sync.v_scrollbar.setValue)
            h2 = table_to_sync.v_scrollbar.valueChanged.connect(self.v_scrollbar.setValue)
            h3 = self.v_header.sectionResized.connect(_table_2_resized)
            h4 = table_to_sync.v_header.sectionResized.connect(_table_1_resized)
        else:
            h1 = self.h_scrollbar.valueChanged.connect(table_to_sync.h_scrollbar.setValue)
            h2 = table_to_sync.h_scrollbar.valueChanged.connect(self.h_scrollbar.setValue)
            h3 = self.h_header.sectionResized.connect(_table_2_resized)
            h4 = table_to_sync.h_header.sectionResized.connect(_table_1_resized)
        return [h1, h2, h3, h4]

    def get_visible_section_span(
        self, orientation: constants.OrientationStr | constants.Orientation
    ) -> tuple[int, int]:
        """Get range of visible sections.

        Arguments:
            orientation: Whether to get section span for horizontal or vertical header.
        """
        orientation = constants.ORIENTATION.get_enum_value(orientation)
        top_left = core.QPoint(0, 0)
        bottom_right = self.viewport().rect().bottomRight()
        if orientation == constants.HORIZONTAL:
            start = self.indexAt(top_left).column()
            count = self.model().columnCount()
            end = self.indexAt(bottom_right).column()
        else:
            start = self.indexAt(top_left).row()
            count = self.model().rowCount()
            end = self.indexAt(bottom_right).row()
        if count == 0:
            return (-1, -1)
        end = count if end == -1 else end + 1
        return (start, end)


class AbstractItemView(AbstractItemViewMixin, widgets.QAbstractItemView):
    """The basic functionality for item view classes."""


if __name__ == "__main__":
    import pandas as pd

    df = pd.DataFrame(dict(a=[1]))
    app = widgets.app()
    table = widgets.TableView()
    table.set_model_for(df)
    table.show()
    app.exec()
